import re
import warnings
from contextlib import suppress
from datetime import time
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union
from onetick.py.backports import Literal

from onetick import py as otp
from onetick.py import types as ott
from onetick.py import utils
from onetick.py.core.column import _Column
from onetick.py.core.column_operations.base import _Operation
from onetick.py.core.eval_query import _QueryEvalWrapper
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def if_else(self: 'Source', condition: _Operation, if_expr, else_expr) -> 'otp.Column':
    """
    Shortcut for :meth:`~onetick.py.Source.apply` with lambda if-else expression

    Parameters
    ----------
    condition: :class:`Operation`
        - condition for matching ticks

    if_expr: :class:`Operation`, value
        - value or `Operation` to set if `condition` is true

    else_expr: :class:`Operation`, value
        - value or `Operation` to set if `condition` is false

    Returns
    -------
    Column

    Examples
    --------
    Basic example of apply if-else to a tick flow:

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data['Y'] = data.if_else(data['X'] > 2, 1, 0)
    >>> otp.run(data)
                         Time  X  Y
    0 2003-12-01 00:00:00.000  1  0
    1 2003-12-01 00:00:00.001  2  0
    2 2003-12-01 00:00:00.002  3  1

    You can also set column value via :class:`Operation`:

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data['Y'] = data.if_else(data['X'] > 2, data['X'] * 2, 0)
    >>> otp.run(data)
                         Time  X  Y
    0 2003-12-01 00:00:00.000  1  0
    1 2003-12-01 00:00:00.001  2  0
    2 2003-12-01 00:00:00.002  3  6

    See Also
    --------
    :py:meth:`onetick.py.Source.apply`
    """
    return self.apply(lambda tick: if_expr if condition else else_expr)


def where_clause(
    self: 'Source', condition, discard_on_match: bool = False, stop_on_first_mismatch: bool = False
) -> Tuple['Source', 'Source']:
    """
    Split source in two branches depending on ``condition``:
    one branch with ticks that meet the condition and
    the other branch with ticks that don't meet the condition.

    Original source object is not modified.

    Parameters
    ----------
    condition: :class:`Operation`, :func:`eval`
        Condition expression to filter ticks or object evaluating another query.
        In the latter case another query should have only one tick as a result with only one field.
    discard_on_match: bool
        Inverts the ``condition``.

        Ticks that don't meet the condition will be returned in the first branch,
        and ticks that meet the condition will be returned in the second branch.
    stop_on_first_mismatch: bool
        If set, no ticks will be propagated in the first branch
        starting with the first tick that does not meet the ``condition``.

        Other branch will contain all ticks starting with the first mismatch, even if they don't meet the condition.

    See Also
    --------
    | :meth:`Source.where`
    | :meth:`Source.__getitem__`
    | **WHERE_CLAUSE** OneTick event processor

    Examples
    --------

    Filtering based on expression:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> odd, even = data.where_clause(data['X'] % 2 == 1)
    >>> otp.run(odd)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.002  3
    >>> otp.run(even)
                         Time  X
    0 2003-12-01 00:00:00.001  2
    1 2003-12-01 00:00:00.003  4

    Filtering based on the result of another query:

    >>> another_query = otp.Tick(WHERE='mod(X, 2) = 1')
    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data, _ = data.where_clause(otp.eval(another_query))
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.002  3

    Using ``discard_on_match`` parameter to invert the condition:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> even, odd = data.where_clause(data['X'] % 2 == 1, discard_on_match=True)
    >>> otp.run(even)
                         Time  X
    0 2003-12-01 00:00:00.001  2
    1 2003-12-01 00:00:00.003  4
    >>> otp.run(odd)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.002  3

    Using ``stop_on_first_mismatch`` parameter to not propagate ticks after first mismatch:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data, other = data.where_clause(data['X'] % 2 == 1, stop_on_first_mismatch=True)
    >>> otp.run(data)
            Time  X
    0 2003-12-01  1

    But other branch will contain all ticks after the mismatch, even if they don't meet the condition:

    >>> otp.run(other)
                         Time  X
    0 2003-12-01 00:00:00.001  2
    1 2003-12-01 00:00:00.002  3
    2 2003-12-01 00:00:00.003  4
    """
    if not isinstance(condition, (_Operation, _QueryEvalWrapper)):
        raise TypeError(f"Unsupported type of value for 'condition' parameter: {type(condition)}")

    if isinstance(condition, _Operation):
        condition = condition._make_python_way_bool_expression()
    if isinstance(condition, _QueryEvalWrapper):
        condition = condition.to_eval_string(self._tmp_otq)
    where_branch = self.copy(
        ep=otq.WhereClause(
            where=str(condition), discard_on_match=discard_on_match, stop_on_first_mismatch=stop_on_first_mismatch
        )
    )

    if_source = where_branch.copy()
    if_source.node().out_pin("IF")

    else_source = where_branch.copy()
    else_source.node().out_pin("ELSE")
    # TODO: add ability to remove then this ep, because it is required only for right output
    else_source.sink(otq.Passthrough())

    return if_source, else_source


def where(self: 'Source', condition, discard_on_match: bool = False, stop_on_first_mismatch: bool = False) -> 'Source':
    """
    Filter ticks that meet the ``condition``.

    Returns new object, original source object is not modified.

    Parameters
    ----------
    condition: :class:`Operation`, :func:`eval`
        Condition expression to filter ticks or object evaluating another query.
        In the latter case another query should have only one tick as a result with only one field.

    discard_on_match: bool
        Inverts the ``condition``.

        Ticks that don't meet the condition will be returned.

    stop_on_first_mismatch: bool
        If set, no ticks will be propagated starting with the first tick that does not meet the ``condition``.

    See Also
    --------
    | :meth:`Source.where_clause`
    | :meth:`Source.__getitem__`
    | **WHERE_CLAUSE** OneTick event processor

    Examples
    --------

    Filtering based on expression:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.where(data['X'] % 2 == 1)
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.002  3

    Filtering based on the result of another query:

    >>> another_query = otp.Tick(WHERE='mod(X, 2) = 1')
    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.where(otp.eval(another_query))
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.002  3

    Using ``discard_on_match`` parameter to invert the condition:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.where(data['X'] % 2 == 1, discard_on_match=True)
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.001  2
    1 2003-12-01 00:00:00.003  4

    Using ``stop_on_first_mismatch`` parameter to not propagate ticks after first mismatch:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.where(data['X'] % 2 == 1, stop_on_first_mismatch=True)
    >>> otp.run(data)
            Time  X
    0 2003-12-01  1
    """
    return self.where_clause(
        condition, discard_on_match=discard_on_match, stop_on_first_mismatch=stop_on_first_mismatch
    )[0]


def _get_integer_slice(self: 'Source', item: slice) -> Optional['Source']:
    """
    Treat otp.Source object as a sequence of ticks
    and apply common python integer slicing logic to it.
    """
    start, stop, step = item.start, item.stop, item.step
    for v in (start, stop, step):
        if v is not None and not isinstance(v, int):
            return None

    # let's filter out cases that we don't want to support
    if step is not None and step <= 0:
        raise ValueError("step value can't be negative or zero")
    if stop is not None and stop == 0:
        raise ValueError("stop value can't be zero")
    # pylint: disable=chained-comparison
    if start and stop and start > 0 and stop > 0 and start >= stop:
        raise ValueError("stop value can't be less than start")
    if start and start < 0 and stop and stop > 0:
        raise ValueError("start value can't be negative when start value is positive")

    def add_counter(src, force=False):
        if '__NUM__' not in src.schema or force:
            if '__NUM__' in src.schema:
                src = src.drop('__NUM__')
            src = src.agg({'__NUM__': otp.agg.count()}, running=True, all_fields=True)
        return src

    result = self.copy()
    if start:
        if start > 0:
            result = add_counter(result)
            result, _ = result[result['__NUM__'] > start]
        if start < 0:
            result = result.last(-start)
    if stop:
        if stop > 0:
            result = add_counter(result)
            result, _ = result[result['__NUM__'] <= stop]
        if stop < 0:
            result = add_counter(result)
            last_ticks = result.last(-stop)
            last_ticks['__FLAG__'] = 1
            last_ticks = last_ticks[['__FLAG__', '__NUM__']]
            result = otp.join(
                result, last_ticks, on=result['__NUM__'] == last_ticks['__NUM__'], how='left_outer', rprefix='RIGHT'
            )
            result, _ = result[result['__FLAG__'] == 0]
            result = result.drop(['__FLAG__', 'RIGHT___NUM__'])
    if step:
        if step > 0:  # NOSONAR
            # resetting counter
            result = add_counter(result, force=True)
            result, _ = result[(result['__NUM__'] - 1) % step == 0]
    if '__NUM__' in result.schema:
        result = result.drop('__NUM__')
    return result


def __getitem__(self: 'Source', item):
    """
    Allows to express multiple things:

    - access a field by name

    - filter ticks by condition

    - select subset of fields

    - set order of fields

    Parameters
    ----------
    item: str, :class:`Operation`, :func:`eval`, list of str

        - ``str`` is to access column by name or columns specified by regex.

        - ``Operation`` to express filter condition.

        - ``otp.eval`` to express filter condition based on external query

        - ``List[str]`` select subset of specified columns or columns specified in regexes.

        - ``slice[List[str]::]`` set order of columns

        - ``slice[Tuple[str, Type]::]`` type defaulting

        - ``slice[:]`` alias to :meth:`Source.copy()`

        - ``slice[int:int:int]`` select ticks the same way as elements in python lists

    Returns
    -------
    Column, Source or tuple of Sources
        - Column if column name was specified.

        - Two sources if filtering expression or eval was provided: the first one is for ticks that pass condition
            and the second one that do not.

    Examples
    --------

    Access to the `X` column: add `Y` based on `X`

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data['Y'] = data['X'] * 2
    >>> otp.run(data)
                         Time  X  Y
    0 2003-12-01 00:00:00.000  1  2
    1 2003-12-01 00:00:00.001  2  4
    2 2003-12-01 00:00:00.002  3  6

    Filtering based on expression:

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data_more, data_less = data[(data['X'] > 2)]
    >>> otp.run(data_more)
                         Time  X
    0 2003-12-01 00:00:00.002  3
    >>> otp.run(data_less)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2

    Filtering based on the result of another query. Another query should
    have only one tick as a result with only one field (whatever it names).

    >>> exp_to_select = otp.Ticks(WHERE=['X > 2'])
    >>> data = otp.Ticks(X=[1, 2, 3], Y=['a', 'b', 'c'], Z=[.4, .3, .1])
    >>> data, _ = data[otp.eval(exp_to_select)]
    >>> otp.run(data)
                         Time  X  Y    Z
    0 2003-12-01 00:00:00.002  3  c  0.1

    Select subset of specified columns:

    >>> data = otp.Ticks(X=[1, 2, 3], Y=['a', 'b', 'c'], Z=[.4, .3, .1])
    >>> data = data[['X', 'Z']]
    >>> otp.run(data)
                         Time  X    Z
    0 2003-12-01 00:00:00.000  1  0.4
    1 2003-12-01 00:00:00.001  2  0.3
    2 2003-12-01 00:00:00.002  3  0.1

    Slice with list will keep all columns, but change order:

    >>> data=otp.Tick(Y=1, X=2, Z=3)
    >>> otp.run(data)
            Time  Y  X  Z
    0 2003-12-01  1  2  3
    >>> data = data[['X', 'Y']:]
    >>> otp.run(data)
            Time  X  Y  Z
    0 2003-12-01  2  1  3

    Slice can be used as short-cut for :meth:`Source.copy`:

    >>> data[:] # doctest: +ELLIPSIS
    <onetick.py.sources.ticks.Tick object at ...>

    Slices can use integers.
    In this case ticks are selected the same way as elements in python lists.

    >>> data = otp.Ticks({'A': [1, 2, 3, 4, 5]})

    Select first 3 ticks:

    >>> otp.run(data[:3])
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3

    Skip first 3 ticks:

    >>> otp.run(data[3:])
                         Time  A
    0 2003-12-01 00:00:00.003  4
    1 2003-12-01 00:00:00.004  5

    Select last 3 ticks:

    >>> otp.run(data[-3:])
                         Time  A
    0 2003-12-01 00:00:00.002  3
    1 2003-12-01 00:00:00.003  4
    2 2003-12-01 00:00:00.004  5

    Skip last 3 ticks:

    >>> otp.run(data[:-3])
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2

    Skip first and last tick:

    >>> otp.run(data[1:-1])
                         Time  A
    0 2003-12-01 00:00:00.001  2
    1 2003-12-01 00:00:00.002  3
    2 2003-12-01 00:00:00.003  4

    Select every second tick:

    >>> otp.run(data[::2])
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.002  3
    2 2003-12-01 00:00:00.004  5

    Select every second tick, not including first and last tick:

    >>> otp.run(data[1:-1:2])
                         Time  A
    0 2003-12-01 00:00:00.001  2
    1 2003-12-01 00:00:00.003  4

    Regular expressions can be used to select fields too:

    >>> data = otp.Tick(A=1, AA=2, AB=3, B=4, BB=5, BA=6)
    >>> otp.run(data['A.*'])
            Time  A  AA  AB  BA
    0 2003-12-01  1   2   3   6

    Note that by default pattern is matched in any position of the string.
    Use characters ^ and $ to specify start and end of the string:

    >>> otp.run(data['^A'])
            Time  A  AA  AB
    0 2003-12-01  1   2   3

    Several regular expressions can be specified too:

    >>> otp.run(data[['^A+$', '^B+$']])
            Time  A  AA  B  BB
    0 2003-12-01  1   2  4   5

    See Also
    --------
    | :meth:`Source.table`: another and more generic way to select subset of specified columns
    | **PASSTHROUGH** OneTick event processor
    | **WHERE_CLAUSE** OneTick event processor

    """

    strict = True

    with suppress(TypeError):
        return self.where_clause(item)

    if isinstance(item, slice):

        result = self._get_integer_slice(item)
        if result:
            return result

        if item.step:
            raise AttributeError("Source columns slice with step set makes no sense")
        if item.start and item.stop:
            raise AttributeError("Source columns slice with both start and stop set is not available now")
        if not item.start and item.stop:
            raise AttributeError("Source columns slice with only stop set is not implemented yet")
        if item.start is None and item.stop is None:
            return self.copy()

        item = item.start
        strict = False

        if isinstance(item, tuple):
            item = dict([item])

        elif isinstance(item, list):
            if not item:
                return self.copy()
            item_type = list(set([type(x) for x in item]))

            if len(item_type) > 1:
                raise AttributeError(f"Different types {item_type} in slice list is not supported")
            if item_type[0] == tuple:
                item = dict(item)

    if isinstance(item, (list, str)):
        # check if item has regex characters
        item_list = [item] if isinstance(item, str) else item
        try:
            items_to_passthrough, use_regex = self._columns_names_regex(item_list)
        except TypeError:
            use_regex = False
        if use_regex:
            src = self.copy()
            src.sink(otq.Passthrough(fields=','.join(items_to_passthrough), use_regex=True))
            return src

    if isinstance(item, list):
        # ---------
        # TABLE
        # ---------
        items = []

        for it in item:
            if isinstance(it, _Column):
                items.append(it.name)
            elif isinstance(it, str):
                items.append(it)
            else:
                raise ValueError(f"It is not supported to filter '{it}' object of '{type(it)}' type")

        # validation
        for item in items:
            if item not in self.schema:
                existing_columns = ", ".join(self.schema.keys())
                raise AttributeError(f"There is no '{item}' column. There are existing columns: {existing_columns}")

        columns = {
            column_name: self.schema[column_name] for column_name in items if not self._check_key_is_meta(column_name)
        }

        return self.table(strict=strict, **columns)

    if isinstance(item, dict):
        return self.table(strict=strict, **item)

    # way to set type
    if isinstance(item, tuple):
        name, dtype = item
        warnings.warn(
            'Using tuple with name and type in otp.Source.__getitem__() is not supported anymore,'
            ' change your code to use otp.Source.schema object instead.',
            FutureWarning,
        )
        return self._set_field_by_tuple(name, dtype)

    name = item
    if name not in self.__dict__:
        raise KeyError(
            f'Column name {name} is not in the schema. Please, check that this column '
            'is in the schema or add it using the .schema property'
        )
    if not isinstance(self.__dict__[name], _Column):
        raise AttributeError(f"There is no '{name}' column")
    return self.__dict__[name]


@inplace_operation
def dropna(
    self: 'Source', how: Literal["any", "all"] = "any", subset: Optional[List[Any]] = None, inplace=False
) -> Optional['Source']:
    """
    Drops ticks that contain NaN values according to the policy in the ``how`` parameter

    Parameters
    ----------
    how: "any" or "all"

        ``any`` - filters out ticks if at least one field has NaN value

        ``all`` - filters out ticks if all fields have NaN values.
    subset: list of str
        list of columns to check for NaN values. If ``None`` then all columns are checked.
    inplace: bool
        the flag controls whether operation should be applied inplace.

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    Drop ticks where **at least one** field has ``nan`` value.

    >>> data = otp.Ticks([[     'X',     'Y'],
    ...                   [     0.0,     1.0],
    ...                   [ otp.nan,     2.0],
    ...                   [     4.0, otp.nan],
    ...                   [ otp.nan, otp.nan],
    ...                   [     6.0,    7.0]])
    >>> data = data.dropna()
    >>> otp.run(data)[['X', 'Y']]
        X   Y
    0 0.0 1.0
    1 6.0 7.0

    Drop ticks where **all** fields have ``nan`` values.

    >>> data = otp.Ticks([[     'X',     'Y'],
    ...                   [     0.0,     1.0],
    ...                   [ otp.nan,     2.0],
    ...                   [     4.0, otp.nan],
    ...                   [ otp.nan, otp.nan],
    ...                   [     6.0,    7.0]])
    >>> data = data.dropna(how='all')
    >>> otp.run(data)[['X', 'Y']]
        X   Y
    0 0.0 1.0
    1 NaN 2.0
    2 4.0 NaN
    3 6.0 7.0

    Drop ticks where **all** fields in **subset** of columns have ``nan`` values.

    >>> data = otp.Ticks([[     'X',     'Y',    'Z'],
    ...                   [     0.0,     1.0, otp.nan],
    ...                   [ otp.nan,     2.0, otp.nan],
    ...                   [     4.0, otp.nan, otp.nan],
    ...                   [ otp.nan, otp.nan, otp.nan],
    ...                   [     6.0,     7.0, otp.nan]])
    >>> data = data.dropna(how='all', subset=['X', 'Y'])
    >>> otp.run(data)[['X', 'Y', 'Z']]
        X   Y   Z
    0 0.0 1.0 NaN
    1 NaN 2.0 NaN
    2 4.0 NaN NaN
    3 6.0 7.0 NaN

    """
    if how not in ["any", "all"]:
        raise ValueError(f"It is expected to see 'any' or 'all' values for 'how' parameter, but got '{how}'")

    condition = None
    columns = self.columns(skip_meta_fields=True)
    if subset is not None:
        for column_name in subset:
            if column_name not in columns:
                raise ValueError(f"There is no '{column_name}' column in the source")
            if columns[column_name] is not float:
                raise ValueError(f"Column '{column_name}' is not float type")

    for column_name, dtype in columns.items():
        if subset is not None and column_name not in subset:
            continue
        if dtype is float:
            if condition is None:
                condition = self[column_name] != ott.nan
            else:
                if how == "any":
                    condition &= self[column_name] != ott.nan
                elif how == "all":
                    condition |= self[column_name] != ott.nan

    self.sink(otq.WhereClause(where=str(condition)))
    return self


@inplace_operation
def time_filter(
    self: 'Source',
    discard_on_match: bool = False,
    start_time: Union[str, int, time] = 0,
    end_time: Union[str, int, time] = 0,
    day_patterns: Union[str, List[str]] = "",
    timezone=utils.default,  # type: ignore
    end_time_tick_matches: bool = False,
    inplace=False,
) -> Optional['Source']:
    """
    Filters ticks by time.

    Parameters
    ----------
    discard_on_match : bool, optional
        If ``True``, then ticks that match the filter will be discarded.
        Otherwise, only ticks that match the filter will be passed.
    start_time : str or int or :py:class:`datetime.time`, optional
        Start time of the filter, string must be in the format ``HHMMSSmmm``.
        Default value is 0.
    end_time : str or int or :py:class:`datetime.time`, optional
        End time of the filter, string must be in the format ``HHMMSSmmm``.
        To filter ticks for an entire day, this parameter should be set to 240000000.
        Default value is 0.
    day_patterns : list or str
        Pattern or list of patterns that determines days for which the ticks can be propagated.
        A tick can be propagated if its date matches one or more of the patterns.
        Three supported pattern formats are:

        1. ``month.week.weekdays``, 0 month means any month, 0 week means any week,
            6 week means the last week of the month for a given weekday(s),
            weekdays are digits for each day, 0 being Sunday.

        2. ``month/day``, 0 month means any month.

        3. ``year/month/day``, 0 year means any year, 0 month means any month.

    timezone : str, optional
        Timezone of the filter.
        Default value is ``configuration.config.tz``
        or timezone set in the parameter of :py:func:`onetick.py.run`.
    end_time_tick_matches : bool, optional
        If ``True``, then the end time is inclusive.
        Otherwise, the end time is exclusive.
    inplace : bool, optional
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object. Default value is ``False``.

    Returns
    -------
    :class:`Source` or ``None``
        Returns ``None`` if ``inplace=True``.

    See also
    --------
    **TIME_FILTER** OneTick event processor

    Examples
    --------
    >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL')
    >>> data = data.time_filter(start_time='000000001', end_time='000000003')
    >>> otp.run(data, start=otp.dt(2022, 3, 1), end=otp.dt(2022, 3, 2))
                         Time  PRICE  SIZE
    0 2022-03-01 00:00:00.001    1.4    10
    1 2022-03-01 00:00:00.002    1.4    50

    """
    if timezone is utils.default:
        # doesn't work without expr for some reason
        timezone = 'expr(_TIMEZONE)'

    if day_patterns:
        if isinstance(day_patterns, str):
            day_patterns = [day_patterns]
        for day_pattern in day_patterns:
            if not re.match(r"(^\d\d?\.[0-6].\d\d?$)|(^\d\d?\/\d\d?$)|(^\d{1,4}\/\d\d?\/\d\d?$)", day_pattern):
                raise ValueError(f"Invalid day pattern: {day_pattern}")

    if isinstance(start_time, time):
        start_time = start_time.strftime('%H%M%S%f')[:-3]

    if isinstance(end_time, time):
        end_time = end_time.strftime('%H%M%S%f')[:-3]

    day_patterns = ",".join(day_patterns)
    self.sink(
        otq.TimeFilter(
            discard_on_match=discard_on_match,
            start_time=start_time,
            end_time=end_time,
            timezone=timezone,
            day_patterns=day_patterns,
            end_time_tick_matches=end_time_tick_matches,
        )
    )
    return self


@inplace_operation
def skip_bad_tick(
    self: 'Source',
    field: Union[str, _Column],
    discard_on_match: bool = False,
    jump_threshold: float = 2.0,
    num_neighbor_ticks: int = 5,
    use_absolute_values: bool = False,
    inplace=False,
) -> Optional['Source']:
    """
    Discards ticks based on whether the value of the attribute specified by ``field`` differs from the value
    of the same attribute in the surrounding ticks more times than a given threshold.
    Uses SKIP_BAD_TICK EP.

    Parameters
    ----------
    field: str, :py:class:`~onetick.py.Column`
        Name of the field (must be present in the input tick descriptor).
    discard_on_match: bool
        When set to ``True`` only ticks that did not match the filter are propagated,
        otherwise ticks that satisfy the filter condition are propagated.
    jump_threshold: float
        A threshold to determine if a tick is "good" or "bad."

        Good ticks are the ticks whose ``field`` value differs less than ``jump_threshold`` times
        from the ``field``'s value of less than or half of the surrounding ``num_neighbor_ticks`` ticks.
    num_neighbor_ticks: int
        The number of ticks before this tick and after this tick to compare a tick against.
    use_absolute_values: bool
        When set to ``True``, use absolute values of numbers when checking whether they are within the jump threshold.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    **SKIP_BAD_TICK** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------
    Keep ticks whose price did not jump by more than 20% relative to the surrounding ticks:

    >>> data = otp.Ticks(X=[10, 11, 15, 11, 9, 10])
    >>> data = data.skip_bad_tick(field="X", jump_threshold=1.2, num_neighbor_ticks=1)
    >>> otp.run(data)
                         Time   X
    0 2003-12-01 00:00:00.000  10
    1 2003-12-01 00:00:00.001  11
    2 2003-12-01 00:00:00.003  11
    3 2003-12-01 00:00:00.005  10

    Same example, but with passing column as ``field`` parameter:

    >>> data = otp.Ticks(X=[10, 11, 15, 11, 9, 10])
    >>> data = data.skip_bad_tick(field=data["X"], jump_threshold=1.2, num_neighbor_ticks=1)
    >>> otp.run(data)
                         Time   X
    0 2003-12-01 00:00:00.000  10
    1 2003-12-01 00:00:00.001  11
    2 2003-12-01 00:00:00.003  11
    3 2003-12-01 00:00:00.005  10

    If you want to keep only "bad ticks", which don't match the filter,
    set ``discard_on_match`` parameter to ``True``:

    >>> data = otp.Ticks(X=[10, 11, 15, 11, 9, 10])
    >>> data = data.skip_bad_tick(field=data["X"], jump_threshold=1.2, num_neighbor_ticks=1, discard_on_match=True)
    >>> otp.run(data)
                         Time   X
    0 2003-12-01 00:00:00.002  15
    1 2003-12-01 00:00:00.004   9

    In case, if you need to compare values on an absolute basis, set ``use_absolute_values`` parameter to ``True``:

    >>> data = otp.Ticks(X=[10, -11, -15, 11, 9, 10])
    >>> data = data.skip_bad_tick(field=data["X"], jump_threshold=1.2, num_neighbor_ticks=1, use_absolute_values=True)
    >>> otp.run(data)
                         Time   X
    0 2003-12-01 00:00:00.000  10
    1 2003-12-01 00:00:00.001  -11
    2 2003-12-01 00:00:00.003  11
    3 2003-12-01 00:00:00.005  10
    """
    if isinstance(field, _Column):
        field = field.name

    if field not in self.schema:
        raise ValueError(f'Field {field} not in the schema.')

    self.sink(otq.SkipBadTick(
        discard_on_match=discard_on_match,
        jump_threshold=jump_threshold,
        field=field,
        num_neighbor_ticks=num_neighbor_ticks,
        use_absolute_values=use_absolute_values,
    ))

    return self


@inplace_operation
def character_present(
    self: 'Source',
    field: Union[str, _Column],
    characters: Union[str, List[str]],
    characters_field: Union[str, _Column] = "",
    discard_on_match: bool = False,
    inplace: bool = False,
):
    """
    Propagates ticks based on whether the value of the field specified by `field` contains a character
    in the set of characters specified by `characters`.
    Uses **CHARACTER_PRESENT** EP.

    Parameters
    ----------
    field: str, :py:class:`~onetick.py.Column`
        Name of the field (must be present in the input tick descriptor).
    characters: str, List[str]
        A set of characters that are searched for in the value of the `field`.
        If set as string, works as list of characters.
    characters_field: str, :py:class:`~onetick.py.Column`
        If specified, will take a current value of that field and append it to `characters`, if any.
    discard_on_match: bool
        When set to ``True`` only ticks that did not match the filter are propagated,
        otherwise ticks that satisfy the filter condition are propagated.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.

    See also
    --------
    **CHARACTER_PRESENT** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    Select ticks that have the N or T in EXCHANGE field

    >>> data = otp.DataSource('TEST_DATABASE', tick_type='TRD', symbols='A')  # doctest: +SKIP
    >>> data = data[['PRICE', 'SIZE', 'EXCHANGE']]  # doctest: +SKIP
    >>> data = data.character_present(field=data['EXCHANGE'], characters='NT')  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
                         Time  PRICE   SIZE EXCHANGE
    0 2003-12-01 00:00:00.000  28.44  55100        N
    1 2003-12-01 00:00:00.001  28.44    100        T
    2 2003-12-01 00:00:00.002  28.44    200        T
    3 2003-12-01 00:00:00.003  28.45    100        T
    4 2003-12-01 00:00:00.004  28.44    500        T

    Select ticks that have the N or T in EXCHANGE field and character set in OLD_EXCHANGE field

    >>> data = otp.DataSource('TEST_DATABASE', tick_type='TRD', symbols='A')  # doctest: +SKIP
    >>> data = data.character_present(  # doctest: +SKIP
    ...     field=data['EXCHANGE'], characters='NT', characters_field=data['OLD_EXCHANGE'],
    ... )
    >>> data = data[['PRICE', 'SIZE', 'EXCHANGE']]  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
                         Time  PRICE   SIZE EXCHANGE
    0 2003-12-01 00:00:00.000  28.44  55100        N
    1 2003-12-01 00:00:00.001  28.44    100        B
    2 2003-12-01 00:00:00.002  28.44    200        B
    3 2003-12-01 00:00:00.003  28.45    100        T
    4 2003-12-01 00:00:00.004  28.44    200        T
    """
    if isinstance(field, _Column):
        field = field.name

    if isinstance(characters_field, _Column):
        characters_field = characters_field.name

    if isinstance(characters, list):
        characters = ''.join(characters)

    for name, value in zip(['field', 'characters_field'], [field, characters_field]):
        if not value:
            continue

        if value not in self.schema:
            raise ValueError(f'Field {value}, passed as parameter `{name}`, not in the schema.')

        if not (self.schema[value] == str or issubclass(self.schema[value], otp.string)):
            raise TypeError(
                f'Field {value}, passed as parameter `{name}`, has incompatible type: {self.schema[value]}, '
                f'expected: str',
            )

    self.sink(otq.CharacterPresent(
        field=field,
        characters=characters,
        characters_field=characters_field,
        discard_on_match=discard_on_match,
    ))

    return self


def primary_exch(self: 'Source', discard_on_match: bool = False) -> Tuple['Source', 'Source']:
    """
    Propagates the tick if its exchange is the PRIMARY exchange of the security. The primary exchange information
    is supplied through the Reference Database. It expects the security level symbol (IBM, not IBM.N) and works
    by looking for a field called ``EXCHANGE`` and filtering out ticks where the field does not match
    the primary exchange for the security.

    .. note::
        This EP may not work correctly with OneTick Cloud databases, due to differences
        in format of exchange names in RefDB and in tick data.

    Parameters
    ----------
    discard_on_match: bool
        When set to ``True`` only ticks from non-primary exchange are propagated,
        otherwise ticks from primary exchange are propagated.

    See also
    --------
    **PRIMARY_EXCH** OneTick event processor

    Returns
    -------
    Two :class:`Source` for each of if-else branches

    Examples
    --------

    Get ticks from primary exchange:

    >>> src = otp.DataSource('SOME_DB', tick_type='TRD', symbols='AAA', date=otp.date(2003, 12, 1))  # doctest: +SKIP
    >>> src, _ = src.primary_exch()  # doctest: +SKIP
    >>> otp.run(src, symbol_date=otp.date(2003, 12, 1))  # doctest: +SKIP
                         Time  PRICE  SIZE EXCHANGE
    0 2003-12-01 00:00:00.001   26.5   150        B
    1 2003-12-01 00:00:00.002   25.7    20        B

    Get all ticks, but mark ticks from primary exchange in column ``T``:

    >>> src = otp.DataSource('SOME_DB', tick_type='TRD', symbols='AAA', date=otp.date(2003, 12, 1))  # doctest: +SKIP
    >>> primary, other = src.primary_exch()  # doctest: +SKIP
    >>> primary['T'] = 1  # doctest: +SKIP
    >>> other['T'] = 0  # doctest: +SKIP
    >>> data = otp.merge([primary, other])  # doctest: +SKIP
    >>> otp.run(src, symbol_date=otp.date(2003, 12, 1))  # doctest: +SKIP
                         Time  PRICE  SIZE EXCHANGE  T
    0 2003-12-01 00:00:00.000   25.2   100        A  0
    1 2003-12-01 00:00:00.001   26.5   150        B  1
    2 2003-12-01 00:00:00.002   25.7    20        B  1
    3 2003-12-01 00:00:00.003   24.8    40        A  0
    """
    source = self.copy(ep=otq.PrimaryExch(discard_on_match=discard_on_match))

    if_source = source.copy()
    if_source.node().out_pin("IF")

    else_source = source.copy()
    else_source.node().out_pin("ELSE")

    return if_source, else_source



import datetime

import pandas as pd

from onetick.py.backports import cached_property
from onetick.py import utils
from . import CallbackBase


class LogCallback(CallbackBase):
    def __init__(self, query_name):
        print(f'Running query {query_name}')
        super().__init__()

    def process_symbol_name(self, symbol_name):
        print(f'Processing symbol {symbol_name}')


class ManualDataframeCallback(CallbackBase):
    """
    This callback class can be used to generate the same :pandas:`pandas.DataFrame` result as in otp.run.
    Unlike otp.run, here result is constructed manually, one tick at a time.
    This may lead to lower memory usage in some cases.
    See task PY-863 for details.
    """

    def __init__(self, timezone, callback_objects=None):
        # getting timezone from user as string or from replicating as object
        self._timezone = utils.tz.get_tzfile_by_name(timezone) if isinstance(timezone, str) else timezone
        # list of all ManualDataframeCallback objects, passed when replicating
        self._callback_objects = [] if callback_objects is None else callback_objects
        self._callback_objects.append(self)
        # ManualDataframeCallback is replicated for each symbol
        self._symbol_name = None
        # dict of columns names and lists with their values
        self._columns = None
        self._defaults = None
        self._result = None

    def replicate(self):
        """
        Object is replicated for each symbol.
        Passing timezone and list to save all objects to aggregate them in the end.
        """
        return ManualDataframeCallback(self._timezone, callback_objects=self._callback_objects)

    def process_symbol_name(self, symbol_name):
        """
        Called once for each ManualDataframeCallback object
        """
        self._symbol_name = symbol_name

    def process_tick_descriptor(self, tick_descriptor):
        """
        Called for each change of schema.
        For example, different dates in database may have different schemas.
        In this case we are merging them together.
        """
        if self._columns is None:
            self._columns = {'Time': []}
            self._defaults = {}
        for name, type_dict in tick_descriptor:
            if name not in self._columns:
                dtype_str = type_dict['type']
                default = self._get_default_value_by_type(dtype_str)
                # save default value for column, will be used in other places
                self._defaults[name] = default
                # update previous ticks with default value of the new field
                self._columns[name] = [default] * len(self._columns['Time'])

    def process_tick(self, tick, time):
        """
        Called for each tick of the symbol.
        """
        assert self._columns is not None
        for name in self._columns:
            if name == 'Time':
                value = self._convert_to_timezone(time)
            elif name not in tick:
                value = self._defaults[name]
            else:
                value = tick[name]
                if isinstance(value, datetime.datetime):
                    # datetime values are always in UTC
                    value = self._convert_to_timezone(value)
            self._columns[name].append(value)

    def done(self):
        """
        Called once for each symbol after all ticks are processed.
        """
        self._result = pd.DataFrame(self._columns)

    @cached_property
    def result(self):
        """
        Must be called by the user after callback object is used in otp.run.
        Aggregating results from all ManualDataframeCallback objects
        and returning pandas dataframe or dict of such (same as result of otp.run in 'df' output mode).
        """
        results = {
            callback_object._symbol_name: callback_object._result
            for callback_object in self._callback_objects
            if callback_object._result is not None
        }
        # if we have only one symbol, return dataframe, not dict
        if len(results) == 1:
            return results.popitem()[1]
        return results

    def _convert_to_timezone(self, dt):
        """
        Converting timezone-naive ``dt`` object in UTC timezone
        to the timezone specified by user and returning also timezone-naive object.
        """
        return utils.convert_timezone(dt, 'UTC', self._timezone)

    def _get_default_value_by_type(self, dtype_str):
        """
        Converting type names returned by process_tick_descriptor()
        """
        defaults = {
            'int': 0,
            'float': 0.0,
            'string': '',
            'datetime': self._convert_to_timezone(
                # epoch in timezone specified by user converted to UTC, that's what OneTick would return
                pd.Timestamp(1970, 1, 1).tz_localize(self._timezone).tz_convert('UTC').tz_localize(None)
            )
        }
        return defaults[dtype_str]



from abc import ABC, abstractmethod
from typing import List
from copy import deepcopy


class _NodeRule(ABC):
    key_params: List[str] = []

    @abstractmethod
    def build(self, eps, print_out=False):
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}({id(self)})'


class _Assign(_NodeRule):
    key_params = ['key']

    def __init__(self, ep, key):
        self.ep = ep
        self.key = key

    def build(self, eps, print_out=False):
        if self.key not in eps:
            eps[self.key] = self.ep.copy()

        if print_out:
            print(f'[API] Create "{str(eps[self.key]).strip()}" (id={id(eps[self.key])})')

        return eps[self.key]


class _TickType(_NodeRule):
    key_params = ['key']

    def __init__(self, ep, key, tt):
        self.ep = ep
        self.key = key
        self.tt = tt

    def build(self, eps, print_out=False):
        if self.key not in eps:
            eps[self.key] = self.ep.copy()

        if print_out:
            print(f'[API] Set tick type "{self.tt}" for "{str(eps[self.key]).strip()}" (id={id(eps[self.key])})')

        eps[self.key].tick_type(self.tt)
        return eps[self.key]


class _Symbol(_NodeRule):
    key_params = ['key']

    def __init__(self, ep, key, symbol):
        self.ep = ep
        self.key = key
        self.symbol = symbol

    def build(self, eps, print_out=False):
        if self.key not in eps:
            eps[self.key] = self.ep.copy()

        if print_out:
            print(f'[API] Set symbol "{self.symbol}" for "{str(eps[self.key]).strip()}" (id={id(eps[self.key])})')

        eps[self.key].set_symbol(self.symbol)
        return eps[self.key]


class _NodeName(_NodeRule):
    key_params = ['key']

    def __init__(self, key, name):
        self.key = key
        self.name = name

    def build(self, eps, print_out=False):
        eps[self.key].node_name(self.name)

        if print_out:
            print(f'[API] Set name "{self.name}" for "{str(eps[self.key]).strip()}" (id={id(eps[self.key])})')

        return eps[self.key]


class _Sink(_NodeRule):
    key_params = ['p_key', 'key']

    def __init__(self, p_ep, p_key, p_out_pin, ep, key, in_pin):
        self.p_ep = p_ep
        self.p_key = p_key
        self.p_out_pin = p_out_pin
        self.ep = ep
        self.key = key
        self.in_pin = in_pin

    def build(self, eps, print_out=False):
        if self.p_key not in eps:
            eps[self.p_key] = self.p_ep.copy()
        if self.key not in eps:
            eps[self.key] = self.ep.copy()

        t_p_ep = eps[self.p_key]
        t_ep = eps[self.key]

        if self.p_out_pin:
            t_p_ep = eps[self.p_key][self.p_out_pin]
        if self.in_pin:
            t_ep = eps[self.key](self.in_pin)

        if print_out:
            print(f'[API] Connect "{str(t_p_ep).strip()}"[out_pin={self.p_out_pin}] -> '
                  f'"{str(t_ep).strip()}"[self.in_pin={self.in_pin}] (id={id(t_p_ep)} and id={id(t_ep)})')

        t_p_ep.sink(t_ep)
        return t_ep


class _Source(_NodeRule):
    key_params = ['p_key', 'key']

    def __init__(self, p_ep, p_key, p_in_pin, ep, key, out_pin):
        self.p_ep = p_ep
        self.p_key = p_key
        self.p_in_pin = p_in_pin
        self.ep = ep
        self.key = key
        self.out_pin = out_pin

    def build(self, eps, print_out=False):
        if self.p_key not in eps:
            eps[self.p_key] = self.p_ep.copy()
        if self.key not in eps:
            eps[self.key] = self.ep.copy()

        t_p_ep = eps[self.p_key]
        t_ep = eps[self.key]

        if self.p_in_pin:
            t_p_ep = eps[self.p_key](self.p_in_pin)
        if self.out_pin:
            t_ep = eps[self.key][self.out_pin]

        if print_out:
            print(f'[API] Connect "{str(t_ep).strip()}"[out_pin={self.out_pin}] -> '
                  f'"{str(t_p_ep).strip()}"[in_pin={self.p_in_pin}] (id={id(t_ep)} and id={id(t_p_ep)})')

        t_p_ep.source(t_ep)
        return t_p_ep


class _SourceByKey(_NodeRule):
    key_params = ['p_key', 'key']

    def __init__(self, p_key, ep, key, out_pin):
        self.p_key = p_key
        self.ep = ep
        self.key = key
        self.out_pin = out_pin

    def build(self, eps, print_out=False):
        if self.key not in eps:
            eps[self.key] = self.ep.copy()

        t_p_ep = eps[self.p_key]
        t_ep = eps[self.key]

        if self.out_pin:
            t_ep = eps[self.key][self.out_pin]

        if print_out:
            print(f'[API] Connect "{str(t_ep).strip()}" -> "{str(t_p_ep).strip()}" (id={id(t_ep)} and id={id(t_p_ep)})')

        t_p_ep.source(t_ep)
        return t_p_ep


class _NodesHistory:
    def __init__(self):
        self._rules: List[_NodeRule] = []

    def assign(self, ep, key):
        self._rules.append(_Assign(ep, key))

    def tick_type(self, ep, key, tt):
        self._rules.append(_TickType(ep, key, tt))

    def symbol(self, ep, key, symbol):
        self._rules.append(_Symbol(ep, key, symbol))

    def node_name(self, key, name):
        self._rules.append(_NodeName(key, name))

    def sink(self, p_ep, p_key, p_out_pin, ep, key, in_pin):
        self._rules.append(_Sink(p_ep, p_key, p_out_pin, ep, key, in_pin))

    def source(self, p_ep, p_key, p_in_pin, ep, key, out_pin):
        self._rules.append(_Source(p_ep, p_key, p_in_pin, ep, key, out_pin))

    def source_by_key(self, p_key, ep, key, out_pin):
        self._rules.append(_SourceByKey(p_key, ep, key, out_pin))

    def rebuild(self, keys: dict):
        """Rebuild history, change uuid for each node"""
        processed = set()
        for rule in self._rules:
            if id(rule) in processed:
                # this can happen when we are merging the branch of the source with this source.
                # In this case some rules in the source and the branch will be the same.
                # (e.g. in dump() function)
                continue
            processed.add(id(rule))
            for key_param in rule.key_params:
                key = getattr(rule, key_param)
                setattr(rule, key_param, keys[key])

    def build(self, eps, root_key, print_out=False):
        if print_out:
            print("")
            print("[[API OUTPUT STARTS]]")

        if print_out:
            print('number of rules:', len(self._rules))
        for rule in self._rules:
            rule.build(eps, print_out=print_out)

        if print_out:
            print("[[API OUTPUT ENDS]]")

        res = eps[root_key]

        assert res is not None

        return res

    def copy(self, deep=False):
        return list(self._rules) if not deep else deepcopy(self._rules)

    def add(self, other_rules):
        processed = set(id(rule) for rule in self._rules)
        for rule in other_rules:
            if id(rule) in processed:
                # this can happen when we are merging the branch of the source with this source.
                # In this case some rules in the source and the branch will be the same.
                # (e.g. in dump() function)
                continue
            processed.add(id(rule))
            self._rules.append(rule)



import datetime as dt

import onetick.py as otp
from onetick.py.otq import otq

import onetick.py.core._source
import onetick.py.functions
import onetick.py.db._inspection
from onetick.py.core.source import Source

from .. import utils, configuration

from .common import update_node_tick_type


class Empty(Source):
    """
    Empty data source

    Parameters
    ----------
    db: str
        Name of the database from which to take schema.
    symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
        Symbol(s) from which data should be taken.
    tick_type: str,
        Name of the tick_type from which to take schema.
    start, end: :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`, \
                    :py:class:`onetick.py.adaptive`
        Time interval from which the data should be taken.
    schema: dict
        Schema to use in case db and/or tick_type are not set.
    kwargs:
        Deprecated. Use ``schema`` instead.
        Schema to use in case db and/or tick_type are not set.

    Examples
    --------
    We can define schema:

    >>> data = otp.Empty(schema={'A': str, 'B': int})
    >>> otp.run(data)
    Empty DataFrame
    Columns: [A, B, Time]
    Index: []
    >>> data.schema
    {'A': <class 'str'>, 'B': <class 'int'>}

    Or we can get schema from the database:

    >>> data = otp.Empty(db='SOME_DB', tick_type='TT')
    >>> data.schema
    {'X': <class 'int'>}
    """

    def __init__(
        self,
        db=utils.adaptive_to_default,
        symbol=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        schema=None,
        **kwargs,
    ):
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        schema = self._select_schema(schema, kwargs)

        columns = {}

        if (tick_type is not utils.adaptive and
                db != configuration.config.get('default_db') and db is not utils.adaptive_to_default):
            try:
                db_obj = onetick.py.db._inspection.DB(db)
                params = {'tick_type': tick_type}
                if end is not utils.adaptive:
                    params['end'] = end
                columns = db_obj.schema(**params)
            except Exception:
                pass  # do not raise an exception if no data found, because it is empty _source and does not matter

        else:
            columns = schema

        super().__init__(
            _symbols=symbol, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(db), schema=columns,
        )

    def base_ep(self, db):
        src = Source(otq.TickGenerator(fields="long ___NOTHING___=0"))
        update_node_tick_type(src, 'TICK_GENERATOR', db)
        return src



from collections import defaultdict

from onetick.py.core._internal._nodes_history import _NodesHistory


class _ProxyNode:
    """
    This class wraps node in _Source with goal to track changes
    """

    def __init__(self, ep, key, _, out_pin, refresh_func=None):
        self._ep = ep
        self._key = key
        self._out_pin = out_pin
        self._name = ""
        self._refresh_func = refresh_func

        self._hist = _NodesHistory()
        self._hist.assign(ep, key)

    def sink(self, p_out_pin, ep, key, in_pin, _, move_node=True):
        """
        Connect self._ep[p_out_pin|self._out_pin] -> ep[in_pin]
        """
        self._refresh_func()

        if not p_out_pin:
            p_out_pin = self._out_pin

        self._hist.sink(self._ep, self._key, p_out_pin, ep, key, in_pin)

        t_p_ep = self._ep
        if p_out_pin:
            t_p_ep = t_p_ep[p_out_pin]

        t_ep = ep
        if in_pin:
            t_ep = ep

        if move_node:
            t_p_ep.sink(t_ep)
            self._ep = ep
            self._key = key
            #  it is not the first, then no need to store this pin
            self._out_pin = None
            self._name = ""

        return self._ep

    def source(self, p_in_pin, ep, key, _, out_pin):
        """
        Connect self._ep[p_in_pin] <- ep[out_pin]
        """
        self._refresh_func()

        self._hist.source(self._ep, self._key, p_in_pin, ep, key, out_pin)

        t_p_ep = self._ep
        if p_in_pin:
            t_p_ep = t_p_ep(p_in_pin)

        t_ep = ep
        if out_pin:
            t_ep = t_ep[out_pin]

        t_p_ep.source(t_ep)

        return self._ep

    def source_by_key(self, to_key, ep, key, _, out_pin):
        """
        Connect node with key=to_key <- ep[out_pin]
        """
        self._refresh_func()

        self._hist.source_by_key(to_key, ep, key, out_pin)

        return self._ep

    def node_name(self, name=None, key=None):
        if name is not None:
            if key:
                # set node_name by key
                self._hist.node_name(key, name)
            else:
                self._hist.node_name(self._key, name)
                self._name = name
            return None
        else:
            return self._name

    def tick_type(self, tt):
        self._refresh_func()
        self._hist.tick_type(self._ep, self._key, tt)
        self._ep.tick_type(tt)

    def symbol(self, symbol):
        self._refresh_func()
        self._hist.symbol(self._ep, self._key, symbol)
        self._ep.set_symbol(symbol)

    def get(self):
        return self._ep

    def key(self, _key=None):
        if _key:
            self._key = _key

        return self._key

    def out_pin(self, _out_pin=None):
        if _out_pin:
            self._out_pin = _out_pin

        return self._out_pin

    def copy_graph(self, eps=None, print_out=False):
        if eps is None:
            eps = defaultdict()

        return self._hist.build(eps, self._key, print_out=print_out), self._key, None, self._out_pin

    def rebuild_graph(self, keys: dict):
        """
        Changing all uuids in this node and it's history.
        Need it for making a deep copy of node.

        Parameters
        ----------
        keys: dict
            Mapping from old key to new key
        """
        self._key = keys[self._key]
        self._hist.rebuild(keys)

    def copy_rules(self, deep=False):
        return self._hist.copy(deep=deep)

    def add_rules(self, rules):
        self._hist.add(rules)



from typing import TYPE_CHECKING

from onetick import py as otp

if TYPE_CHECKING:
    import pandas
    from onetick.py.core.source import Source


def plot(self: 'Source', y, x='Time', kind='line', **kwargs):
    """
    Executes the query with known properties and builds a plot resulting dataframe.

    Uses the :pandas:`pandas.DataFrame.plot` method to plot data.
    Other parameters could be specified through the ``kwargs``.

    Parameters
    ----------
    x: str
        Column name for the X axis
    y: str
        Column name for the Y axis
    kind: str
        The kind of plot

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data.plot(y='X', kind='bar')  # doctest: +SKIP
    """
    result = self.copy()
    return result[[y, x]]().plot(x=x, y=y, kind=kind, **kwargs)


def count(self: 'Source', **kwargs) -> int:
    """
    Returns the number of ticks in the query.

    Adds an aggregation that calculate total ticks count, and *executes a query*.
    Result is a single value -- number of ticks. Possible application is the Jupyter when
    a developer wants to check data presences for example.

    Parameters
    ----------
    kwargs
        parameters that will be passed to :py:func:`otp.run <onetick.py.run>`

    Returns
    -------
    int

    See Also
    --------
    | :py:func:`onetick.py.agg.count`
    | :py:func:`otp.run <onetick.py.run>`
    | :py:meth:`onetick.py.Source.head`
    | :py:meth:`onetick.py.Source.tail`

    Examples
    --------

    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data.count()
    3

    >>> data = otp.Empty()
    >>> data.count()
    0
    """
    result = self.copy()
    df = otp.run(result.agg({'__num_rows': otp.agg.count()}), **kwargs)
    if df.empty:
        return 0
    return int(df['__num_rows'][0])


def head(self: 'Source', n=5, **kwargs) -> 'pandas.DataFrame':
    """
    *Executes the query* and returns first ``n`` ticks as a pandas dataframe.

    It is useful in the Jupyter case when you want to observe first ``n`` values.

    Parameters
    ----------
    n: int, default=5
        number of ticks to return
    kwargs:
        parameters will be passed to :py:func:`otp.run <onetick.py.run>`

    Returns
    -------
    :pandas:`DataFrame <pandas.DataFrame>`

    See Also
    --------
    | :py:func:`onetick.py.agg.first`
    | :py:func:`otp.run <onetick.py.run>`
    | :py:meth:`onetick.py.Source.tail`
    | :py:meth:`onetick.py.Source.count`

    Examples
    --------

    >>> data = otp.Ticks(X=list('abcdefgik'))
    >>> data.head()[['X']]
        X
    0 a
    1 b
    2 c
    3 d
    4 e
    """
    result = self.copy()
    result = result.first(n=n)  # pylint: disable=E1123
    return otp.run(result, **kwargs)


def tail(self: 'Source', n=5, **kwargs) -> 'pandas.DataFrame':
    """
    *Executes the query* and returns last ``n`` ticks as a pandas dataframe.

    It is useful in the Jupyter case when you want to observe last ``n`` values.

    Parameters
    ----------
    n: int
        number of ticks to return
    kwargs:
        parameters will be passed to :py:func:`otp.run <onetick.py.run>`

    Returns
    -------
    :pandas:`DataFrame <pandas.DataFrame>`

    See Also
    --------
    | :py:func:`onetick.py.agg.last`
    | :py:func:`otp.run <onetick.py.run>`
    | :py:meth:`onetick.py.Source.head`
    | :py:meth:`onetick.py.Source.count`

    Examples
    --------
    >>> data = otp.Ticks(X=list('abcdefgik'))
    >>> data.tail()[['X']]
        X
    0 e
    1 f
    2 g
    3 i
    4 k
    """
    result = self.copy()
    result = result.last(n=n)  # pylint: disable=E1123
    return otp.run(result, **kwargs)



from typing import Any, Callable, List, Optional, Tuple, Union
from types import FunctionType
from datetime import datetime

import onetick.py as otp
from onetick.py.otq import otq
from onetick.py.core.source import Source
from onetick.py.sources.common import update_node_tick_type
from onetick.py.sources.cache import process_otq_params
from onetick.py.types import datetime as otp_datetime

from . import configuration


def _check_status(df):
    if "STATUS" not in df:
        raise RuntimeError("Empty response returned")

    status = df["STATUS"][0]
    if status != "SUCCESS":
        raise RuntimeError(f"Error status returned: {status}")


def _convert_dt_to_str(dt: Union[str, datetime, otp_datetime]):
    if isinstance(dt, str):
        return dt
    elif isinstance(dt, (datetime, otp_datetime)):
        return dt.strftime('%Y%m%d%H%M%S.%f')
    raise ValueError(f"Unsupported value for 'dt' parameter: {type(dt)}")


def _convert_time_intervals(
    time_intervals_to_cache: List[Tuple[Union[str, datetime, otp_datetime], Union[str, datetime, otp_datetime]]],
):
    return "\n".join(
        [
            f"{_convert_dt_to_str(start_time)},{_convert_dt_to_str(end_time)}"
            for start_time, end_time in time_intervals_to_cache
        ]
    )


def create_cache(
    cache_name: str,
    query: Union['Source', Callable, str, None] = None,
    inheritability: bool = True,
    otq_params: Union[dict, None] = None,
    time_granularity: int = 0,
    time_granularity_units: Optional[str] = None,
    timezone: str = "",
    time_intervals_to_cache: Optional[List[tuple]] = None,
    allow_delete_to_everyone: bool = False,
    allow_update_to_everyone: bool = False,
    allow_search_to_everyone: bool = True,
    cache_expiration_interval: int = 0,
    tick_type: str = "ANY",
    symbol: Optional[str] = None,
    db: Optional[str] = None,
):
    """
    Create cache via CREATE_CACHE EP

    If :py:class:`onetick.py.Source` or callable passed as ``query`` parameter,
    cache will be created only for current session.

    Also, this case or passing absolute path to .otq as ``query`` parameter are supported
    only for caching on local OneTick server.

    Parameters ``symbol``, ``db`` and ``tick_type`` could be omitted if you want to use
    default symbol, default database and ``ANY`` tick type.

    Parameters
    ----------
    cache_name: str
        Name of the cache to be created.
    query: :py:class:`onetick.py.Source`, callable, str
        Query to be cached or path to .otq file (in ``filename.otq::QueryName`` format).
        Only queries residing on the server that runs the caching event processors are currently supported.
        For local OneTick server you can pass absolute path to local .otq file.
    inheritability: bool
        Indicates whether results can be obtained by combining time intervals that were cached with intervals
        freshly computed to obtain results for larger intervals.
    otq_params: dict
        OTQ params of the query to be cached.
        Setting `otq_params` in :py:class:`onetick.py.ReadCache` may not override this `otq_params`
        if cache not invalidated.
    time_granularity: int
        Value N for seconds/days/months granularity means that start and end time of the query have to be on N
        second/day/month boundaries relative to start of the day/month/year.
        This doesn't affect the frequency of data within the cache, just the start and end dates.
    time_granularity_units: str, None
        Units used in ``time_granularity`` parameter. Possible values: 'none', 'days', 'months', 'seconds' or None.
    timezone: str
        Timezone of the query to be cached.
    time_intervals_to_cache: List[tuple]
        List of tuples with start and end times in ``[(<start_time_1>, <end_time_1>), ...]`` format,
        where ``<start_time>`` and ``<end_time>`` should be one of these:

        * string in ``YYYYMMDDhhmmss[.msec]`` format.
        * :py:class:`datetime.datetime`
        * :py:class:`onetick.py.types.datetime`

        If specified only these time intervals can be cached. Ignored if ``inheritability=True``.
        If you try to make a query outside defined interval, error will be raised.
    allow_delete_to_everyone: bool
        When set to ``True`` everyone is allowed to delete the cache.
    allow_update_to_everyone: bool
        When set to ``True`` everyone is allowed to update the cache.
    allow_search_to_everyone: bool
        When set to ``True`` everyone is allowed to read the cached data.
    cache_expiration_interval: int
        If set to a non-zero value determines the periodicity of cache clearing, in seconds.
        The cache will be cleared every X seconds, triggering new query executions when data is requested.
    tick_type: str
        Tick type.
    symbol: str, list of str, list of otq.Symbol, :py:class:`onetick.py.Source`, :pandas:`pandas.DataFrame`, optional
        ``symbols`` parameter of ``otp.run()``.
    db: str
        Database.

    Note
    ----
    This function does NOT populate the cache, it's only reserves the cache name in the OneTick memory.
    Cache is only populated when an attempt is made to read the data from it via :py:class:`onetick.py.ReadCache`.

    See also
    --------
    | **CREATE_CACHE** OneTick event processor
    | :py:class:`onetick.py.ReadCache`
    | :py:func:`onetick.py.delete_cache`

    Examples
    --------
    Simple cache creation from .otq file on OneTick server under ``OTQ_FILE_PATH``

    >>> otp.create_cache(  # doctest: +SKIP
    ...    cache_name="some_cache", query="CACHE_EXAMPLE.otq::slowquery",
    ...    tick_type="TRD", db="LOCAL",
    ... )

    Cache creation from function

    >>> def query_func():
    ...    return otp.DataSource("COMMON", tick_type="TRD", symbols="AAPL")
    >>> otp.create_cache(
    ...    cache_name="some_cache", query=query_func, tick_type="TRD", db="LOCAL",
    ... )

    Create cache for time intervals with different datetime types:

    >>> otp.create_cache(  # doctest: +SKIP
    ...    cache_name="some_cache",
    ...    query="query_example.otq::query"),
    ...    inheritability=False,
    ...    time_intervals_to_cache=[
    ...        ("20220601123000.000000", "20220601183000.000000"),
    ...        (datetime(2022, 6, 2, 12, 30), datetime(2003, 1, 2, 18, 30)),
    ...        (otp.datetime(2022, 6, 3, 12, 30), otp.datetime(2022, 6, 3, 18, 30)),
    ...    ],
    ...    timezone="GMT",
    ...    tick_type="TRD",
    ...    db="LOCAL",
    ... )

    Create cache with OTQ params:

    >>> otp.create_cache(  # doctest: +SKIP
    ...    cache_name="some_cache",
    ...    query="query_example.otq::query"),
    ...    otq_params={"some_param": "some_value"},
    ...    timezone="GMT",
    ...    tick_type="TRD",
    ...    db="LOCAL",
    ... )
    """
    if query is None:
        raise ValueError("Parameter `query` should be set")

    if time_granularity_units is None:
        time_granularity_units = "none"

    if time_granularity_units not in {'none', 'days', 'months', 'seconds'}:
        raise ValueError(f"Incorrect `time_granularity_units` param value passed: {time_granularity_units}")

    time_granularity_units = time_granularity_units.upper()

    if symbol is None:
        symbol = configuration.config.default_symbol

    if db is None:
        db = configuration.config.default_db

    time_intervals_str = ""
    if time_intervals_to_cache:
        time_intervals_str = _convert_time_intervals(time_intervals_to_cache)

    otq_params_str = process_otq_params(otq_params)

    otq_file_path = None
    if query:
        if isinstance(query, FunctionType):
            query = query()

        if isinstance(query, Source):
            # will create sub-query after source is initialized below
            otq_file_path = 'THIS::create_cache_query'
        elif isinstance(query, str):
            otq_file_path = query
        else:
            raise ValueError(f"Passed `query` parameter value with incorrect type: {type(query)}")

    source = Source(
        otq.CreateCache(
            cache_name=cache_name,
            otq_file_path=otq_file_path,
            inheritability=inheritability,
            otq_params=otq_params_str,
            time_granularity=time_granularity,
            time_granularity_units=time_granularity_units,
            timezone=timezone,
            time_intervals_to_cache=time_intervals_str,
            allow_delete_to_everyone=allow_delete_to_everyone,
            allow_update_to_everyone=allow_update_to_everyone,
            allow_search_to_everyone=allow_search_to_everyone,
            cache_expiration_interval=cache_expiration_interval,
        ),
    )

    if isinstance(query, Source):
        # create temp file with query
        query._store_in_tmp_otq(source._tmp_otq, name='create_cache_query')

    update_node_tick_type(source, tick_type, db)

    res = otp.run(source, symbols=symbol)
    _check_status(res)


def delete_cache(
    cache_name: str,
    apply_to_entire_cache: bool = True,
    per_cache_otq_params: Union[dict, None] = None,
    tick_type: str = "ANY",
    symbol: Optional[str] = None,
    db: Optional[str] = None,
):
    """
    Delete cache via DELETE_CACHE EP

    Parameters
    ----------
    cache_name: str
        Name of the cache to be deleted.
    apply_to_entire_cache: bool
        When set to ``True`` deletes the cache for all symbols and time intervals.
    per_cache_otq_params: dict
        Deletes cache that have been associated with this OTQ parameters during its creation.
        Value of this parameter should be equal to the value of ``otq_params`` of
        :func:`<onetick.py.create_cache>`.
    tick_type: str
        Tick type.
    symbol: str, list of str, list of otq.Symbol, :py:class:`onetick.py.Source`, :pandas:`pandas.DataFrame`, optional
        ``symbols`` parameter of ``otp.run()``.
    db: str
        Database.

    See also
    --------
    | **DELETE_CACHE** OneTick event processor
    | :py:class:`onetick.py.ReadCache`
    | :py:func:`onetick.py.create_cache`

    Examples
    --------
    Simple cache deletion

    >>> otp.delete_cache(  # doctest: +SKIP
    ...    cache_name="some_cache", tick_type="TRD", symbol="SYM", db="LOCAL",
    ... )
    """
    if symbol is None:
        symbol = configuration.config.default_symbol

    if db is None:
        db = configuration.config.default_db

    source = Source(
        otq.DeleteCache(
            cache_name=cache_name,
            apply_to_entire_cache=apply_to_entire_cache,
            per_cache_otq_params=process_otq_params(per_cache_otq_params),
        ),
    )

    update_node_tick_type(source, tick_type, db)

    res = otp.run(source, symbols=symbol)
    _check_status(res)


def modify_cache_config(
    cache_name: str,
    param_name: str,
    param_value: Any,
    tick_type: str = "ANY",
    symbol: Optional[str] = None,
    db: Optional[str] = None,
):
    """
    Modify cache configuration via MODIFY_CACHE_CONFIG EP

    Parameters
    ----------
    cache_name: str
        Name of the cache to be deleted.
    param_name: str
        The name of the configuration parameter to be changed.
        Supported parameters:

        * ``inheritability``
        * ``time_granularity``
        * ``time_granularity_units``
        * ``timezone``
        * ``allow_search_to_everyone``
        * ``allow_delete_to_everyone``
        * ``allow_update_to_everyone``

    param_value: Any
        New value of configuration parameter. Will be converted to string.
    tick_type: str
        Tick type.
    symbol: str, list of str, list of otq.Symbol, :py:class:`onetick.py.Source`, :pandas:`pandas.DataFrame`, optional
        ``symbols`` parameter of ``otp.run()``.
    db: str
        Database.

    See also
    --------
    | **MODIFY_CACHE_CONFIG** OneTick event processor
    | :py:class:`onetick.py.ReadCache`
    | :py:func:`onetick.py.create_cache`

    Examples
    --------
    Simple cache config modification

    >>> otp.modify_cache_config(  # doctest: +SKIP
    ...    cache_name="some_cache",
    ...    param_name="time_granularity",
    ...    param_value=3,
    ...    tick_type="TRD", symbol="SYM", db="LOCAL",
    ... )
    """
    if symbol is None:
        symbol = configuration.config.default_symbol

    if db is None:
        db = configuration.config.default_db

    param_name = param_name.upper()
    param_value = str(param_value)

    source = Source(
        otq.ModifyCacheConfig(
            cache_name=cache_name,
            config_parameter_name=param_name,
            config_parameter_value=param_value,
        ),
    )

    update_node_tick_type(source, tick_type, db)

    res = otp.run(source, symbols=symbol)
    _check_status(res)



from collections import UserDict
from functools import partial

from onetick.py import types as ott
from onetick.py.core.column_operations._methods._internal import MethodResult
from onetick.py.core.column_operations._methods.op_types import are_ints_not_time

NSECTIME_TO_STR_FORMAT = "%Y-%m-%d %H:%M:%S.%J"
NSECTIME_TO_STR_FORMAT_2 = "%Y/%m/%d %H:%M:%S.%J"
MSECTIME_TO_STR_FORMAT = "%Y-%m-%d %H:%M:%S.%q"


def float_to_int(prev_op, dtype=int):
    # CASE should be uppercased because it can be used in per-tick script
    op_str = f"CASE({str(prev_op)} > 0, 1, floor({str(prev_op)}), ceil({str(prev_op)}))"
    return MethodResult(op_str, dtype)


def float_to_str(prev_op, dtype=str):
    precision = getattr(prev_op, "_precision", 7)

    op_str = f"tostring({str(prev_op)}, 10, {precision})"

    if precision == 0:
        # otherwise 1.7 would be 1, but should be 1.0, because it is still float TODO: forbid it?
        op_str = f"({op_str} + '.0')"
    # if

    # we need it to remove trailing zeros
    op_str = fr'regex_replace({op_str}, "(\.\d+?)0+\b", "\1")'
    return MethodResult(op_str, dtype)


def num_to_decimal(prev_op):
    return MethodResult(f'decimal({str(prev_op)})', ott.decimal)


def decimal_to_str(prev_op, dtype=str):
    precision = getattr(prev_op, "_precision", 8)
    op_str = f"decimal_to_string({str(prev_op)}, {precision})"
    return MethodResult(op_str, dtype)


def decimal_to_float(prev_op):
    return MethodResult(f'{str(prev_op)}', float)


def nsectime_to_str(prev_op, dtype=str):
    op_str = f"nsectime_format('{NSECTIME_TO_STR_FORMAT}', {str(prev_op)}, _TIMEZONE)"
    return MethodResult(op_str, dtype)


def msectime_to_str(prev_op, dtype=str):
    op_str = f"time_format('{MSECTIME_TO_STR_FORMAT}', {str(prev_op)}, _TIMEZONE)"
    return MethodResult(op_str, dtype)


def nsectime_to_int(prev_op, dtype=int):
    op_str = f"nsectime_to_long({str(prev_op)})"
    return MethodResult(op_str, dtype)


def msectime_to_int(prev_op, dtype=int):
    op_str = str(prev_op)
    return MethodResult(op_str, dtype)


def nsectime_to_msectime(prev_op):
    op_str = f"nsectime({str(prev_op)})"
    return MethodResult(op_str, ott.msectime)


def msectime_to_nsectime(prev_op):
    op_str = f"nsectime({str(prev_op)})"
    return MethodResult(op_str, ott.nsectime)


def int_to_str(prev_op, dtype=str):
    op_str = f"tostring({str(prev_op)})"
    return MethodResult(op_str, dtype)


def int_to_float(prev_op):
    op_str = f"{str(prev_op)} * 1.0"
    return MethodResult(op_str, float)


def int_to_nsectime(prev_op):
    op_str = f"nsectime({str(prev_op)})"
    return MethodResult(op_str, ott.nsectime)


def int_to_msectime(prev_op):
    op_str = str(prev_op)
    return MethodResult(op_str, ott.msectime)


def str_to_float(prev_op):
    op_str = f"atof({str(prev_op)})"
    return MethodResult(op_str, float)


def str_to_decimal(prev_op):
    op_str = f"string_to_decimal({str(prev_op)})"
    return MethodResult(op_str, ott.decimal)


def str_to_int(prev_op, dtype=int):
    op_str = f"atol({str(prev_op)})"
    return MethodResult(op_str, dtype)


def str_to_nsectime(prev_op):
    op_str = _str_to_time(prev_op)
    return MethodResult(op_str, ott.nsectime)


def str_to_msectime(prev_op):
    op_str = _str_to_time(prev_op)
    return MethodResult(op_str, ott.msectime)


def _str_to_time(prev_op):
    result = str(prev_op)

    first_option = f"parse_nsectime('{NSECTIME_TO_STR_FORMAT}', {result}, _TIMEZONE)"
    second_option = f"parse_nsectime('{NSECTIME_TO_STR_FORMAT_2}', {result}, _TIMEZONE)"

    # CASE should be uppercased because it can be used in per-tick script
    return f"CASE(position('-', {result}) > 0, 1, {first_option}, {second_option})"


def bool_to_int(prev_op, dtype=int):
    op_str = f"CASE({str(prev_op)}, true, 1, 0)"
    return MethodResult(op_str, dtype)


def bool_to_float(prev_op):
    op_str = f"CASE({str(prev_op)}, true, 1.0, 0.0)"
    return MethodResult(op_str, float)


def _same_to_same(prev_op, dtype=None):
    if dtype is None:
        dtype = prev_op.dtype
    return MethodResult(str(prev_op), dtype)


class _ConversionsDict(UserDict):
    def __init__(self, conversions):
        self.data = conversions

    def __getitem__(self, key):
        if not (isinstance(key, tuple) and len(key) == 2):
            raise ValueError("wrong usage of _ConversionsDict")

        from_, to = key
        if from_ == to:
            return _same_to_same

        try:
            return super().__getitem__(key)
        except KeyError:
            pass

        new_from, new_to = None, None

        # int subclasses are converted by the same method
        if are_ints_not_time(from_):
            new_from = int
        if are_ints_not_time(to):
            new_to = int

        # otp.string and str are converted by the same methods
        if from_ is not str and issubclass(from_, str):
            new_from = str
        if to is not str and issubclass(to, str):
            new_to = str

        if new_from is not None or new_to is not None:
            new_key = (new_from or from_, new_to or to)
            if new_key[0] == new_key[1]:
                method = _same_to_same
            else:
                method = super().__getitem__(new_key)
            if issubclass(to, str) or are_ints_not_time(to):
                method = partial(method, dtype=to)
            return method

        raise TypeError(f"can not convert {from_} to {to}")


CONVERSIONS = _ConversionsDict({(float, int): float_to_int,
                                (float, str): float_to_str,
                                (float, ott.decimal): num_to_decimal,
                                (ott.decimal, int): float_to_int,
                                (ott.decimal, str): decimal_to_str,
                                (ott.decimal, float): decimal_to_float,
                                (ott.nsectime, str): nsectime_to_str,
                                (ott.msectime, str): msectime_to_str,
                                (ott.nsectime, int): nsectime_to_int,
                                (ott.msectime, int): msectime_to_int,
                                (ott.nsectime, ott.msectime): nsectime_to_msectime,
                                (ott.msectime, ott.nsectime): msectime_to_nsectime,
                                (int, str): int_to_str,
                                (int, float): int_to_float,
                                (int, ott.nsectime): int_to_nsectime,
                                (int, ott.msectime): int_to_msectime,
                                (int, ott.decimal): num_to_decimal,
                                (str, float): str_to_float,
                                (str, ott.decimal): str_to_decimal,
                                (str, int): str_to_int,
                                (str, ott.nsectime): str_to_nsectime,
                                (str, ott.msectime): str_to_msectime,
                                (bool, int): bool_to_int,
                                (bool, float): bool_to_float})



from onetick.py.core.source import Source

from .. import utils

from .common import _common_passthrough_base_ep


class Orders(Source):
    def __init__(
        self, db="S_ORDERS_FIX", symbol=utils.adaptive, start=utils.adaptive, end=utils.adaptive, schema=None, **kwargs,
    ):
        super().__init__(
            _symbols=symbol, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(db), schema=schema, **kwargs,
        )

        self.schema['ID'] = str
        self.schema['BUY_FLAG'] = int
        self.schema['SIDE'] = str
        self.schema['STATE'] = str
        self.schema['ORDTYPE'] = str
        self.schema['PRICE'] = float
        self.schema['PRICE_FILLED'] = float
        self.schema['QTY'] = int
        self.schema['QTY_FILLED'] = int

    def base_ep(self, db):
        return _common_passthrough_base_ep(db, 'ORDER')


class Quotes(Source):
    def __init__(
        self, db=utils.adaptive_to_default, symbol=utils.adaptive, start=utils.adaptive, end=utils.adaptive,
        schema=None, **kwargs,
    ):
        super().__init__(
            _symbols=symbol, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(db), schema=schema, **kwargs,
        )

        self.schema['ASK_PRICE'] = float
        self.schema['BID_PRICE'] = float
        self.schema['ASK_SIZE'] = int
        self.schema['BID_SIZE'] = int

    def base_ep(self, db):
        return _common_passthrough_base_ep(db, 'QTE')


class Trades(Source):
    """
    Trade source object.
    add 'PRICE' and 'SIZE' fields to schema
    """

    def __init__(
        self, db=utils.adaptive_to_default, symbol=utils.adaptive, date=None, start=utils.adaptive, end=utils.adaptive,
        schema=None, **kwargs,
    ):
        if date:
            start, end = date.start, date.end
        super().__init__(
            _symbols=symbol, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(db), schema=schema, **kwargs,
        )

        self.schema['PRICE'] = float
        self.schema['SIZE'] = int

    def base_ep(self, db):
        return _common_passthrough_base_ep(db, 'TRD')


class NBBO(Source):
    def __init__(
        self, db="TAQ_NBBO", symbol=utils.adaptive, start=utils.adaptive, end=utils.adaptive, schema=None, **kwargs,
    ):
        super().__init__(
            _symbols=symbol, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(db), schema=schema, **kwargs,
        )

        self.schema['ASK_PRICE'] = float
        self.schema['BID_PRICE'] = float
        self.schema['ASK_SIZE'] = int
        self.schema['BID_SIZE'] = int

    def base_ep(self, db):
        return _common_passthrough_base_ep(db, 'NBBO')



import warnings

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source
from onetick.py.core.column_operations.base import Raw, OnetickParameter
from onetick.py.core.eval_query import _QueryEvalWrapper
from onetick.py.core._source.tmp_otq import TmpOtq
from onetick.py.compatibility import is_symbols_prepend_db_name_supported

from .. import types as ott
from .. import utils

from .common import update_node_tick_type


class Symbols(Source):
    """
    Construct a source that returns symbol names from the database.

    The **SYMBOL_NAME** field will contain symbol names.

    Parameters
    ----------
    db: str, :py:func:`eval query <onetick.py.eval>`
        Name of the database where to search symbols.
        By default the database used by :py:func:`otp.run <onetick.py.run>` will be inherited.
    keep_db: bool
        Flag that indicates whether symbols should have a database name prefix in the output.
        If True, symbols are returned in *DB_NAME::SYMBOL_NAME* format.
        Otherwise just symbol names are returned.
    pattern: str
        Usual and special characters can be used to search for symbols.
        Special characters are:

        * ``%`` - any number of any characters (zero too)
        * ``_`` - any single character
        * ``\\`` - used to escape special characters

        For example, if you want symbol name starting with ``NQ``, you should write ``NQ%``.
        If you want symbol name to contain literal ``%`` character, you should write ``NQ\\%``.
        ``\\`` is a special character too, so it need to be escaped too
        if you want symbol name to contain literal backslash, e.g. ``NQ\\\\M23``.
        Default is ``%`` (any symbol name).

    for_tick_type: str
        Fetch only symbols belong to this tick type, if specified.
        Otherwise fetch symbols for all tick types.
    show_tick_type: bool
        Add the **TICK_TYPE** column with the information about tick type
    symbology: str
        The destination symbology for a symbol name translation.
        Translation is performed, if destination symbology is not empty
        and is different from that of the queried database.
    show_original_symbols: bool
        Switches original symbol name propagation as a tick field **ORIGINAL_SYMBOL_NAME**
        if symbol name translation is performed (if parameter ``symbology`` is set).

        .. note::

          If this parameter is set to True, database symbols with missing translations are also propagated.
          In this case **ORIGINAL_SYMBOL_NAME** will be presented, but **SYMBOL_NAME** field will be empty.

    discard_on_match: bool
        If True, then parameter ``pattern`` filters out symbols to return from the database.
    cep_method: str
        The method to be used for extracting database symbols in CEP mode.
        Possible values are:

            * *use_db*: symbols will be extracted from the database with intervals
              specified by the ``poll_frequency`` parameter, and new symbols will be output.
            * *use_cep_adapter*: CEP adapter will be used to retrieve and propagate the symbols with every heartbeat.
            * Default: None, the EP will work the same way as for historical queries,
              i.e. will query the database for symbols once.
    poll_frequency: int
        Specifies the time interval in *minutes* to check the database for new symbols.
        This parameter can be specified only if ``cep_method`` is set to *use_db*.
        The minimum value is 1 minute.
    symbols_to_return: str
        Indicates whether all symbols must be returned or only those which are in the query time range.
        Possible values are:

            * *all_in_db*: All symbols are returned.
            * *with_tick_in_query_range*: Only the symbols which have ticks in the query time range are returned.
              This option is allowed only when ``cep_method`` is set to *use_cep_adapter*.

    _tick_type: str
        Custom tick type for the node of the graph.
        By default "ANY" tick type will be set.
    tick_type: str
        .. attention::

            This parameter is deprecated, use parameter ``_tick_type`` instead.
            Do not confuse this parameter with ``for_tick_type``.
            This parameter is used for low-level customization of OneTick graph nodes and is rarely needed.

    start: :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`
        Custom start time of the query.
        By default the start time used by :py:func:`otp.run <onetick.py.run>` will be inherited.
    end: :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`
        Custom end time of the query.
        By default the start time used by :py:func:`otp.run <onetick.py.run>` will be inherited.
    date: :py:class:`datetime.date`
        Alternative way of setting instead of ``start``/``end`` times.


    Note
    ----
    Additional fields that can be added to Symbols will be converted to symbol parameters

    See also
    --------
    | :ref:`Symbols guide <static/concepts/symbols:Symbols: bound and unbound>`
    | **FIND_DB_SYMBOLS** OneTick event processor

    Examples
    --------

    This class can be used to get a list of all symbols in the database:

    >>> data = otp.Symbols('US_COMP_SAMPLE', date=otp.dt(2024, 2, 1))  # doctest: +SKIP
    >>> otp.run(data)                                                  # doctest: +SKIP
              Time  SYMBOL_NAME
    0   2024-02-01            A
    1   2024-02-01          AAL
    2   2024-02-01         AAPL
    3   2024-02-01         ABBV
    4   2024-02-01         ABNB
    ..         ...          ...
    496 2024-02-01          XYL
    497 2024-02-01          YUM
    498 2024-02-01          ZBH
    499 2024-02-01         ZBRA
    500 2024-02-01          ZTS

    By default database name and time interval will be inherited from :py:func:`otp.run <onetick.py.run>`:

    >>> data = otp.Symbols()                                                # doctest: +SKIP
    >>> otp.run(data, symbols='US_COMP_SAMPLE::', date=otp.dt(2024, 2, 1))  # doctest: +SKIP
              Time  SYMBOL_NAME
    0   2024-02-01            A
    1   2024-02-01          AAL
    2   2024-02-01         AAPL
    ..         ...          ...

    Parameter ``keep_db`` can be used to show database name in the output.
    It is useful when querying symbols for many databases:

    >>> data = otp.Symbols(keep_db=True)                                        # doctest: +SKIP
    >>> data = data.first(2)                                                    # doctest: +SKIP
    >>> data = otp.merge([data], symbols=['US_COMP_SAMPLE::', 'CME_SAMPLE::'])  # doctest: +SKIP
    >>> otp.run(data, date=otp.dt(2024, 2, 1))                                  # doctest: +SKIP
            Time          SYMBOL_NAME
    0 2024-02-01    US_COMP_SAMPLE::A
    1 2024-02-01  US_COMP_SAMPLE::AAL
    2 2024-02-01  CME_SAMPLE::CL\\F25
    3 2024-02-01  CME_SAMPLE::CL\\F26
    ..       ...                  ...

    By default symbols for all tick types are returned.
    You can set parameter ``show_tick_type`` to print the tick type for each symbol:

    >>> data = otp.Symbols('US_COMP_SAMPLE', show_tick_type=True)  # doctest: +SKIP
    >>> otp.run(data, date=otp.dt(2024, 2, 1))                     # doctest: +SKIP
               Time  SYMBOL_NAME  TICK_TYPE
    0    2024-02-01            A        DAY
    1    2024-02-01            A       LULD
    2    2024-02-01            A       NBBO
    3    2024-02-01            A        QTE
    4    2024-02-01            A       STAT
    ..          ...          ...        ...

    Parameter ``for_tick_type`` can be used to specify a single tick type for which to return symbols:

    >>> data = otp.Symbols('US_COMP_SAMPLE', show_tick_type=True, for_tick_type='TRD')  # doctest: +SKIP
    >>> otp.run(data, date=otp.dt(2024, 2, 1))                                          # doctest: +SKIP
              Time  SYMBOL_NAME  TICK_TYPE
    0   2024-02-01            A        TRD
    1   2024-02-01          AAL        TRD
    2   2024-02-01         AAPL        TRD
    3   2024-02-01         ABBV        TRD
    4   2024-02-01         ABNB        TRD
    ..         ...          ...        ...

    Parameter ``pattern`` can be used to specify the pattern to filter symbol names:

    >>> data = otp.Symbols('US_COMP_SAMPLE', show_tick_type=True, for_tick_type='TRD',
    ...                    pattern='AAP_')      # doctest: +SKIP
    >>> otp.run(data, date=otp.dt(2024, 2, 1))  # doctest: +SKIP
            Time SYMBOL_NAME TICK_TYPE
    0 2024-02-01        AAPL       TRD

    Parameter ``discard_on_match`` can be used to use ``pattern`` to filter out symbols instead:

    >>> data = otp.Symbols('US_COMP_SAMPLE', show_tick_type=True, for_tick_type='TRD',
    ...                    pattern='AAP_', discard_on_match=True)  # doctest: +SKIP
    >>> otp.run(data, date=otp.dt(2024, 2, 1))                     # doctest: +SKIP
              Time  SYMBOL_NAME  TICK_TYPE
    0   2024-02-01            A        TRD
    1   2024-02-01          AAL        TRD
    2   2024-02-01         ABBV        TRD
    3   2024-02-01         ABNB        TRD
    4   2024-02-01          ABT        TRD
    ..         ...          ...        ...

    ``otp.Symbols`` object can be used to specify symbols for the main query:

    >>> symbols = otp.Symbols('US_COMP_SAMPLE')                           # doctest: +SKIP
    >>> symbols = symbols.first(3)                                        # doctest: +SKIP
    >>> data = otp.DataSource('US_COMP_SAMPLE', tick_type='TRD')          # doctest: +SKIP
    >>> result = otp.run(data, symbols=symbols, date=otp.dt(2024, 2, 1))  # doctest: +SKIP
    >>> result['AAPL'][['Time', 'PRICE', 'SIZE']]                         # doctest: +SKIP
                                    Time   PRICE  SIZE
    0      2024-02-01 04:00:00.008283417  186.50     6
    1      2024-02-01 04:00:00.008290927  185.59     1
    2      2024-02-01 04:00:00.008291153  185.49   107
    3      2024-02-01 04:00:00.010381671  185.49     1
    4      2024-02-01 04:00:00.011224206  185.50     2
    ..                               ...     ...   ...

    >>> result['AAL'][['Time', 'PRICE', 'SIZE']]                          # doctest: +SKIP
                                    Time  PRICE  SIZE
    0      2024-02-01 04:00:00.097381367  14.33     1
    1      2024-02-01 04:00:00.138908789  14.37     1
    2      2024-02-01 04:00:00.726613365  14.36    10
    3      2024-02-01 04:00:02.195702506  14.36    73
    4      2024-02-01 04:01:55.268302813  14.39     1
    ..                               ...    ...   ...

    Additional fields of the ``otp.Symbols`` can be used in the main query as symbol parameters:

    >>> symbols = otp.Symbols('US_COMP_SAMPLE', show_tick_type=True, for_tick_type='TRD')  # doctest: +SKIP
    >>> symbols['PARAM'] = symbols['SYMBOL_NAME'] + '__' + symbols['TICK_TYPE']            # doctest: +SKIP
    >>> data = otp.DataSource('US_COMP_SAMPLE', tick_type='TRD')                           # doctest: +SKIP
    >>> data = data.first(1)                                                               # doctest: +SKIP
    >>> data['S_PARAM'] = data.Symbol['PARAM', str]                                        # doctest: +SKIP
    >>> data = otp.merge([data], symbols=symbols)                                          # doctest: +SKIP
    >>> data = data[['PRICE', 'SIZE', 'S_PARAM']]                                          # doctest: +SKIP
    >>> otp.run(data, date=otp.dt(2024, 2, 1))                                             # doctest: +SKIP
                                 Time    PRICE  SIZE    S_PARAM
    0   2024-02-01 04:00:00.001974784  193.800     4   HSY__TRD
    1   2024-02-01 04:00:00.003547904   57.810    18   OXY__TRD
    2   2024-02-01 04:00:00.006354688   42.810    30   DVN__TRD
    3   2024-02-01 04:00:00.007310080  165.890     9   WMT__TRD
    4   2024-02-01 04:00:00.007833957   43.170    22  INTC__TRD
    ..                            ...      ...   ...        ...

    Use parameter ``symbology`` to specify different symbology to translate to.
    You can also use parameter ``show_original_symbols`` to print original symbols.
    Note that some symbols may not have a translation in target symbology, so their names will be empty:

    >>> data = otp.Symbols('US_COMP', for_tick_type='TRD',
    ...                    symbology='FGV', show_original_symbols=True)                     # doctest: +SKIP
    >>> otp.run(data, start=otp.dt(2023, 5, 15, 9, 30), end=otp.dt(2023, 5, 15, 9, 30, 1))  # doctest: +SKIP
                         Time   SYMBOL_NAME  ORIGINAL_SYMBOL_NAME
    0     2023-05-15 09:30:00  BBG000C2V3D6            US_COMP::A
    1     2023-05-15 09:30:00  BBG00B3T3HD3           US_COMP::AA
    2     2023-05-15 09:30:00  BBG01B0JRCS6          US_COMP::AAA
    3     2023-05-15 09:30:00  BBG00LPXX872         US_COMP::AAAU
    4     2023-05-15 09:30:00  BBG00YZC2Z91          US_COMP::AAC
    ...                   ...           ...                   ...
    10946 2023-05-15 09:30:00                      US_COMP::ZXIET
    10947 2023-05-15 09:30:00                      US_COMP::ZXZZT
    10948 2023-05-15 09:30:00  BBG019XSYC89         US_COMP::ZYME
    10949 2023-05-15 09:30:00  BBG007BBS8B7         US_COMP::ZYNE
    10950 2023-05-15 09:30:00  BBG000BJBXZ2         US_COMP::ZYXI

    **Escaping special characters in the pattern**

    When using patterns with special character, be aware that python strings ``\\`` is a special character too
    and need to be escaped as well:

    >>> print('back\\\\slash')
    back\\slash

    Pattern ``NQ\\\\M23`` in python should be written as ``NQ\\\\\\\\M23``:

    >>> print('NQ\\\\\\\\M23')
    NQ\\\\M23

    Escaping character ``\\`` in python can be avoided with raw strings:

    >>> print(r'NQ\\\\M23')
    NQ\\\\M23
    """

    _PROPERTIES = Source._PROPERTIES + ["_p_db",
                                        "_p_pattern",
                                        "_p_start",
                                        "_p_end",
                                        "_p_for_tick_type",
                                        "_p_keep_db"]

    def __init__(
        self,
        db=None,
        find_params=None,
        keep_db=False,
        pattern='%',
        for_tick_type=None,
        show_tick_type=False,
        symbology='',
        show_original_symbols=False,
        discard_on_match=None,
        cep_method=None,
        poll_frequency=None,
        symbols_to_return=None,
        tick_type=utils.adaptive,
        _tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        date=None,
        schema=None,
        **kwargs,
    ):
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if isinstance(pattern, OnetickParameter):
            pattern = pattern.parameter_expression

        self._p_db = db
        self._p_pattern = pattern
        self._p_start = start
        self._p_end = end
        self._p_keep_db = keep_db
        self._p_for_tick_type = for_tick_type

        if tick_type is not utils.adaptive:
            warnings.warn("In otp.Symbols parameter 'tick_type' is deprecated."
                          " Previously it was incorrectly interpreted by users as a tick type"
                          " for which symbols in the database will be searched."
                          " Instead right now it sets a tick type for a node in OneTick graph "
                          " (this results in symbols from all tick types returned from this source)."
                          " Use parameter 'for_tick_type' to find the symbols for a particular tick type"
                          " and use parameter '_tick_type' if you want set a tick type for a node in OneTick graph.",
                          FutureWarning, stacklevel=2)

        if date and isinstance(date, (ott.datetime, ott.date)):
            start = date.start
            end = date.end

        _symbol = utils.adaptive
        _tmp_otq = None
        if db:
            if isinstance(db, list):
                _symbol = [f"{str(_db).split(':')[0]}::" for _db in db] # noqa
            elif isinstance(db, _QueryEvalWrapper):
                _tmp_otq = TmpOtq()
                _symbol = db.to_eval_string(tmp_otq=_tmp_otq)
            else:
                _symbol = f"{str(db).split(':')[0]}::"  # noqa

        if find_params is not None:
            warnings.warn("In otp.Symbols parameter 'find_params' is deprecated."
                          " Use named parameters instead.",
                          FutureWarning, stacklevel=2)

        _find_params = find_params if find_params is not None else {}

        _find_params.setdefault('pattern', pattern)
        if for_tick_type:
            _find_params['tick_type_field'] = for_tick_type
        _find_params.setdefault('show_tick_type', show_tick_type)

        _find_params.setdefault('symbology', symbology)
        _find_params.setdefault('show_original_symbols', show_original_symbols)

        if 'prepend_db_name' in _find_params:
            raise ValueError('Use parameter `keep_db` instead of passing `prepend_db_name` in `find_params`')

        if discard_on_match is not None:
            _find_params.setdefault('discard_on_match', discard_on_match)
        if cep_method is not None:
            if not isinstance(cep_method, str) or cep_method not in ('use_cep_adapter', 'use_db'):
                raise ValueError(f"Wrong value for parameter 'cep_method': {cep_method}")
            _find_params.setdefault('cep_method', cep_method.upper())
        if poll_frequency is not None:
            _find_params.setdefault('poll_frequency', poll_frequency)
        if symbols_to_return is not None:
            if not isinstance(symbols_to_return, str) or symbols_to_return not in ('all_in_db',
                                                                                   'with_ticks_in_query_range'):
                raise ValueError(f"Wrong value for parameter 'symbols_to_return': {symbols_to_return}")
            _find_params.setdefault('symbols_to_return', symbols_to_return.upper())

        if tick_type is not utils.adaptive and _tick_type is not utils.adaptive:
            raise ValueError("Parameters 'tick_type' and '_tick_type' can't be set simultaneously")
        elif tick_type is not utils.adaptive:
            ep_tick_type = tick_type
        elif _tick_type is not utils.adaptive:
            ep_tick_type = _tick_type
        else:
            ep_tick_type = utils.adaptive

        super().__init__(
            _symbols=_symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(ep_tick_type=ep_tick_type,
                                               keep_db=keep_db, **_find_params),
        )

        self.schema['SYMBOL_NAME'] = str

        if _find_params['show_tick_type']:
            self.schema['TICK_TYPE'] = str

        if _find_params['symbology'] and _find_params['show_original_symbols']:
            self.schema['ORIGINAL_SYMBOL_NAME'] = str

        if _tmp_otq:
            self._tmp_otq.merge(_tmp_otq)

    def base_ep(self, ep_tick_type, keep_db, **params):
        src = Source(otq.FindDbSymbols(**params))

        update_node_tick_type(src, ep_tick_type)
        src.schema['SYMBOL_NAME'] = str

        if not keep_db:
            src["SYMBOL_NAME"] = src["SYMBOL_NAME"].str.regex_replace('.*::', '')

        return src

    @staticmethod
    def duplicate(obj, db=None):
        return Symbols(db=obj._p_db if db is None else db,
                       pattern=obj._p_pattern,
                       start=obj._p_start,
                       end=obj._p_end,
                       keep_db=obj._p_keep_db,
                       for_tick_type=obj._p_for_tick_type)



import time


INDENTS: list[int] = []


def debug(f):
    def wrapper(*args, **kwargs):
        start = time.time()
        cur_indent = INDENTS[-1] + 1 if INDENTS else 0
        INDENTS.append(cur_indent)
        print(' ' * 4 * cur_indent + f.__name__, 'started', args, kwargs)
        result = f(*args, **kwargs)
        print(' ' * 4 * cur_indent + f.__name__, 'executed in', time.time() - start)
        INDENTS.pop()
        return result
    return wrapper



import getpass
import os

from locator_parser.io import PrintWriter, FileReader, FileWriter
from locator_parser.actions import Get, Add, Delete
from locator_parser.acl import DB, User, Role, parse_acl
from locator_parser.common import apply_actions

from .temp import TmpFile
from .types import default


def if_db_in_acl(acl, db_name):
    get_db = Get()
    get_db.add_where(DB, id=db_name)

    return apply_actions(parse_acl, FileReader(acl), PrintWriter(), [get_db], flush=True)


def add_user_to_acl(acl, user):
    add_user = Add(User(name=user))
    add_user.add_where(Role, name="Admin")

    assert apply_actions(parse_acl, FileReader(acl), FileWriter(acl), [add_user], flush=True)


def remove_user_from_acl(acl, user):
    remove_user = Delete()
    remove_user.add_where(Role, name="Admin")
    remove_user.add_where(User, name=user)

    assert apply_actions(parse_acl, FileReader(acl), FileWriter(acl), [remove_user], flush=True)


def if_user_in_acl(acl, user):
    get_user = Get()
    get_user.add_where(Role, name="Admin")
    get_user.add_where(User, name=user)

    return apply_actions(parse_acl, FileReader(acl), PrintWriter(), [get_user], flush=True)


def tmp_acl(clean_up=default):
    data = []
    data.append("<roles>")
    data.append('<role name="Admin">')
    data.append(f'<user name="{getpass.getuser()}" />')
    data.append('<user name="onetick" />')

    ext_users = os.environ.get("TEST_SESSION_ACL_USERS", None)

    if ext_users:
        for usr in ext_users.split(","):
            data.append(f'<user name="{usr}" />')

    data.append("</role>")
    data.append("</roles>")

    data.append("<databases>")
    data.append("</databases>")

    data.append("<event_processors>")
    data.append("")
    data.append('<ep ID="RELOAD_CONFIG">')
    data.append('<allow role="Admin" />')
    data.append("</ep>")
    data.append("")
    data.append('<ep ID="WRITE_TEXT">')
    data.append('<allow role="Admin" />')
    data.append("</ep>")
    data.append("")
    data.append('<ep ID="COMMAND_EXECUTE">')
    data.append('<allow role="Admin" />')
    data.append("</ep>")
    data.append("")
    data.append('<ep ID="READ_FROM_RAW">')
    data.append('<allow role="Admin" />')
    data.append("</ep>")
    data.append("")
    data.append('<ep ID="SAVE_SNAPSHOT">')
    data.append('<allow role="Admin" />')
    data.append("</ep>")
    data.append("</event_processors>")

    if os.getenv('OTP_WEBAPI_TEST_MODE'):
        tmp_file = TmpFile(name="acl.xml", clean_up=clean_up, force=True)
    else:
        tmp_file = TmpFile(suffix=".acl", clean_up=clean_up)

    with open(tmp_file, "w") as fout:
        fout.write("\n".join(data))

    return tmp_file



from typing import TYPE_CHECKING, Tuple

from onetick import py as otp
from onetick.py.otq import otq

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def split(self: 'Source', expr, cases, default=False) -> Tuple['Source', ...]:
    """
    The method splits data using passed expression ``expr`` for several
    outputs by passed ``cases``. The method is the alias for the :meth:`Source.switch`

    Parameters
    ----------
    expr : Operation
        column or column based expression
    cases : list
        list of values or :py:class:`onetick.py.range` objects to split by
    default : bool
        ``True`` adds the default output

    Returns
    -------
    Outputs according to passed cases, number of outputs is number of cases plus one if ``default=True``

    See also
    --------
    | :meth:`Source.switch`
    | **SWITCH** OneTick event processor

    Examples
    --------
    >>> # OTdirective: snippet-name: Source operations.split;
    >>> data = otp.Ticks(X=[0.33, -5.1, otp.nan, 9.4])
    >>> r1, r2, r3 = data.split(data['X'], [otp.nan, otp.range(0, 100)], default=True)
    >>> otp.run(r1)
                         Time   X
    0 2003-12-01 00:00:00.002 NaN
    >>> otp.run(r2)
                         Time     X
    0 2003-12-01 00:00:00.000  0.33
    1 2003-12-01 00:00:00.003  9.40
    >>> otp.run(r3)
                         Time    X
    0 2003-12-01 00:00:00.001 -5.1

    See Also
    --------
    Source.switch
    :py:class:`onetick.py.range`
    """

    output_num = len(cases)

    # format cases
    def to_str(v):
        if isinstance(v, otp.utils.range):
            return "[" + str(v.start) + "," + str(v.stop) + "]"
        elif isinstance(v, str):
            return '"' + v + '"'
        elif isinstance(v, tuple):
            return ",".join(map(to_str, list(v)))

        return str(v)

    cases = [f"{to_str(cases[inx])}:OUT{inx}" for inx in range(output_num)]

    # create ep
    params = dict(switch=str(expr), cases=";".join(cases))
    if default:
        params["default_output"] = "DEF_OUT"

    switch_branch = self.copy(ep=otq.SwitchEp(**params))

    # construct results
    result = []

    for inx in range(output_num):
        res = switch_branch.copy()
        res.node().out_pin(f"OUT{inx}")
        res.sink(otq.Passthrough())

        result.append(res)

    if default:
        res = switch_branch.copy()
        res.node().out_pin("DEF_OUT")
        res.sink(otq.Passthrough())

        result.append(res)

    return tuple(result)


def switch(self: 'Source', expr, cases, default=False) -> Tuple['Source', ...]:
    """
    The method splits data using passed expression for several
    outputs by passed cases. This method is an alias for
    :meth:`Source.split` method.

    Parameters
    ----------
    expr : Operation
        column or column based expression
    cases : list
        list of values or :py:class:`onetick.py.range` objects to split by
    default : bool
        ``True`` adds the default output

    Returns
    -------
    Outputs according to passed cases, number of outputs is number of cases plus one if ``default=True``

    See also
    --------
    | :meth:`Source.split`
    | **SWITCH** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[0.33, -5.1, otp.nan, 9.4])
    >>> r1, r2, r3 = data.switch(data['X'], [otp.nan, otp.range(0, 100)], default=True)
    >>> otp.run(r1)
                         Time   X
    0 2003-12-01 00:00:00.002 NaN
    >>> otp.run(r2)
                         Time     X
    0 2003-12-01 00:00:00.000  0.33
    1 2003-12-01 00:00:00.003  9.40
    >>> otp.run(r3)
                         Time    X
    0 2003-12-01 00:00:00.001 -5.1

    See Also
    --------
    Source.split
    :py:class:`onetick.py.range`
    """

    return self.split(expr, cases, default)



import copy
import warnings
from typing import Type, Sequence

from onetick.py import types as ott
from onetick.py.core.column_operations import _methods

from onetick.py.core.column_operations._methods.op_types import (
    are_ints_not_time,
    are_time,
    are_floats,
    are_strings
)
from onetick.py.core.column_operations._methods.methods import DatetimeSubtractionWarning


class Expr:
    """
    EP parameter's value can be set to an expression.
    Expressions are evaluated before parameters are actually passed to event processors.

    See also
    --------
    :py:attr:`onetick.py.Operation.expr`
    """
    def __init__(self, operation):
        self.operation = operation

    def __str__(self):
        return f'expr({str(self.operation)})'

    def __repr__(self):
        """onetick.query_webapi calls repr() on the object to get the string representation of the object"""
        # TODO
        from onetick.py.otq import otq
        return otq.onetick_repr(self.__str__())


class Operation:
    """
    :py:class:`~onetick.py.Source` column operation container.

    This is the object you get when applying most operations on :py:class:`~onetick.py.Column`
    or on other operations.
    Eventually you can add a new column using the operation you got or pass it as a parameter
    to some functions.

    Examples
    --------
    >>> t = otp.Tick(A=1)
    >>> t['A']
    Column(A, <class 'int'>)
    >>> t['A'] / 2
    Operation((A) / (2))
    >>> t['B'] = t['A'] / 2
    >>> t['B']
    Column(B, <class 'float'>)
    """
    emulation_enabled = False

    def __init__(self, op_func=None, op_params=None, dtype=None, obj_ref=None, op_str=None):
        self._op_func = op_func
        self._op_params = op_params
        self.obj_ref = obj_ref
        self.__warnings = []
        if op_func:
            if op_str:
                raise ValueError("You should specify either op_func or op_str")
            with warnings.catch_warnings(record=True) as warning_list:
                # we want to raise this warning only in some cases
                # that's why we're catching it and saving for later use
                warnings.simplefilter('always', category=DatetimeSubtractionWarning)
                op_str, dtype = self._evaluate_func()

            for w in warning_list:
                if w.category is DatetimeSubtractionWarning:
                    self.__warnings.append(w)
                else:
                    warnings.warn_explicit(w.message, w.category, w.filename, w.lineno)

        self._op_str = op_str
        self._dtype = dtype

    def __bool__(self):
        if Operation.emulation_enabled:
            # True is default for classes without overriden __bool__
            return True
        raise TypeError('It is not allowed to use compare in if-else and while clauses')

    def __str__(self):
        return self.op_str

    def __repr__(self):
        return f"Operation({str(self)})"

    @property
    def dtype(self):
        """
        Returns the type of the column or operation.

        See also
        --------
        :py:meth:`Source.schema <onetick.py.Source.schema>`

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, C='3')
        >>> t['TIMESTAMP'].dtype
        <class 'onetick.py.types.nsectime'>
        >>> t['A'].dtype
        <class 'int'>
        >>> t['B'].dtype
        <class 'float'>
        >>> t['C'].dtype
        <class 'str'>
        """
        dtype = self._dtype
        if not dtype:
            _, dtype = self._evaluate_func(set_fields=True)
        return dtype

    @property
    def op_str(self):
        for w in self.__warnings:
            warnings.showwarning(w.message, w.category, w.filename, w.lineno)
        op_str = self._op_str
        if not op_str:
            op_str, _ = self._evaluate_func(set_fields=True)
        return op_str

    @property
    def expr(self):
        """
        Get expression to use in EP parameters.

        See also
        --------
        :py:class:`~onetick.py.core.column_operations.base.Expr`
        """
        return Expr(self)

    def round(self, precision=0):
        """
        Rounds input column with specified ``precision``.

        Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
        and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity.

        For values that are exactly half-way between two integers (when the fraction part of value is exactly 0.5),
        the rounding method used here is *upwards*, which returns the bigger number.
        For other rounding methods see :func:`otp.math.round <onetick.py.math.round>` function.

        Parameters
        ----------
        precision: int
            Number from -12 to 12.
            Positive precision is precision after the floating point.
            Negative precision is precision before the floating point.

        See also
        --------
        __round__

        Examples
        --------
        >>> t = otp.Tick(A=1234.5678)
        >>> t['B'] = t['A'].round()
        >>> t['C'] = t['A'].round(2)
        >>> t['D'] = t['A'].round(-2)
        >>> otp.run(t)
                Time          A       B        C       D
        0 2003-12-01  1234.5678  1235.0  1234.57  1200.0

        Returns
        -------
        Operation
        """
        return round(self, precision)

    def map(self, arg, default=None):
        """
        Map values of the column to new values according to the mapping in ``arg``.
        If the value is not in the mapping, it is set to the ``default`` value.
        If ``default`` value is not set, it is set to default value for the column type.

        Parameters
        ----------
        arg: dict
            Mapping from old values to new values.
            All values must have the same type, compatible with the column type.
        default: simple value or Column or Operation
            Default value if no mapping is found in ``arg``.
            By default, it is set to default value for the column type.
            (0 for numbers, empty string for strings, etc.)

        Examples
        --------
        >>> t = otp.Ticks(A=[1, 2, 3, 4, 5])
        >>> t['B'] = t['A'].map({1: 10, 2: 20, 3: 30})
        >>> otp.run(t)
                             Time  A   B
        0 2003-12-01 00:00:00.000  1  10
        1 2003-12-01 00:00:00.001  2  20
        2 2003-12-01 00:00:00.002  3  30
        3 2003-12-01 00:00:00.003  4   0
        4 2003-12-01 00:00:00.004  5   0

        Example with ``default`` parameter set:

        >>> t = otp.Ticks(A=[1, 2, 3, 4, 5])
        >>> t['B'] = t['A'].map({1: 10, 2: 20, 3: 30}, default=-1)
        >>> otp.run(t)
                             Time  A   B
        0 2003-12-01 00:00:00.000  1  10
        1 2003-12-01 00:00:00.001  2  20
        2 2003-12-01 00:00:00.002  3  30
        3 2003-12-01 00:00:00.003  4  -1
        4 2003-12-01 00:00:00.004  5  -1

        ``default`` parameter can also be an :py:class:`~onetick.py.Operation` or :py:class:`~onetick.py.Column`.
        For example, to keep unmapped values:

        >>> t = otp.Ticks(A=[1, 2, 3, 4, 5])
        >>> t['B'] = t['A'].map({1: 10, 2: 20, 3: 30}, default=t['A'])
        >>> otp.run(t)
                             Time  A   B
        0 2003-12-01 00:00:00.000  1  10
        1 2003-12-01 00:00:00.001  2  20
        2 2003-12-01 00:00:00.002  3  30
        3 2003-12-01 00:00:00.003  4   4
        4 2003-12-01 00:00:00.004  5   5

        Returns
        -------
        Operation
        """
        if not isinstance(arg, dict) or not arg:
            raise TypeError("map() argument must be a dict with keys and values to map")

        try:
            values_type = ott.get_type_by_objects(arg.values())
        except TypeError as e:
            raise TypeError("map() argument must be a dict with same types for all values") from e

        if default is not None:
            try:
                default_type = ott.get_type_by_objects([default])
                ott.get_type_by_objects([default_type, values_type])
            except TypeError as e:
                raise TypeError(
                    f"map() default value type {default_type} must be compatible with values type {values_type}"
                ) from e

        try:
            keys_type = ott.get_type_by_objects(arg.keys())
        except TypeError as e:
            raise TypeError("map() argument must be a dict with same types for all keys") from e

        try:
            ott.get_type_by_objects([keys_type, self.dtype])
        except TypeError as e:
            raise TypeError(f"map() keys type {keys_type} must be compatible with column type {self.dtype}") from e

        return _Operation(_methods._map, [self, arg, values_type, default])

    def apply(self, lambda_f):
        """
        Apply function or type to column

        Parameters
        ----------
        lambda_f: type or callable
            if type - will convert column to requested type

            if callable - will translate python code to similar OneTick's CASE expression.
            There are some limitations to which python operators can be used in this callable.
            See :ref:`Python callables parsing guide <python callable parser>` article for details.
            In :ref:`Remote OTP with Ray<ray-remote>` any `Callable` must be decorated with `@otp.remote` decorator,
            see :ref:`Ray usage examples<apply-remote-context>` for details.

        Examples
        --------
        Converting type of the column, e.g. string column to integer:

        >>> data = otp.Ticks({'A': ['1', '2', '3']})
        >>> data['B'] = data['A'].apply(int) + 10   # OTdirective: snippet-name: column operations.type convertation;
        >>> otp.run(data)
                             Time  A   B
        0 2003-12-01 00:00:00.000  1  11
        1 2003-12-01 00:00:00.001  2  12
        2 2003-12-01 00:00:00.002  3  13

        More complicated logic:

        >>> data = otp.Ticks({'A': [-321, 0, 123]})
        >>> data['SIGN'] = data['A'].apply(lambda x: 1 if x > 0 else -1 if x < 0 else 0)
        >>> otp.run(data)
                             Time    A  SIGN
        0 2003-12-01 00:00:00.000 -321    -1
        1 2003-12-01 00:00:00.001    0     0
        2 2003-12-01 00:00:00.002  123     1

        See also
        --------
        :py:meth:`onetick.py.Source.apply`
        :ref:`Python callables parsing guide <python callable parser>`
        """
        if isinstance(lambda_f, Type) and ott.is_type_basic(lambda_f):
            return self._convert_to(lambda_f)

        from onetick.py.core.lambda_object import apply_lambda

        return apply_lambda(lambda_f, self)

    def astype(self, to_type):
        """
        Alias for the :meth:`apply` method with type.

        See also
        --------
        :meth:`apply`

        Examples
        --------
        >>> data = otp.Tick(A=1, B=2.2, C='3.3')
        >>> data['A'] = data['A'].astype(str) + 'A'
        >>> data['B'] = data['B'].astype(int) + 1
        >>> data['C'] = data['C'].astype(float) + 0.1
        >>> otp.run(data)
                Time  B   A    C
        0 2003-12-01  3  1A  3.4
        """
        return self.apply(to_type)

    def isin(self, *items):
        """
        Check if column's value is in ``items``.

        Parameters
        ----------
        items
            Possible string or numeric values to be checked against column's value.
            Such values can be passed as the function arguments
            or as the list of such values in the first function argument.

        Returns
        -------
        Operation
            Returns an :py:class:`onetick.py.Operation` object with the value of 1.0 if
            the column's value was found among ``items`` or 0.0 otherwise.

        See also
        --------
        :py:meth:`Source.__getitem__`

        Examples
        --------

        Passing values as function arguments:

        >>> data = otp.Ticks(A=['a', 'b', 'c'])
        >>> data['B'] = data['A'].isin('a', 'c')
        >>> otp.run(data)
                             Time  A    B
        0 2003-12-01 00:00:00.000  a  1.0
        1 2003-12-01 00:00:00.001  b  0.0
        2 2003-12-01 00:00:00.002  c  1.0

        Passing values as a list:

        >>> data = otp.Ticks(A=['a', 'b', 'c'])
        >>> data['B'] = data['A'].isin(['a', 'c'])
        >>> otp.run(data)
                             Time  A    B
        0 2003-12-01 00:00:00.000  a  1.0
        1 2003-12-01 00:00:00.001  b  0.0
        2 2003-12-01 00:00:00.002  c  1.0

        This function's result can be used as filter expression:

        >>> data = otp.Ticks(A=[1, 2, 3, 0])
        >>> yes, no = data[data["A"].isin(0, 1)]    # OTdirective: snippet-name: column operations.is in.constant;
        >>> otp.run(yes)[["A"]]
           A
        0  1
        1  0

        :py:class:`Columns <onetick.py.Column>` and :py:class:`operations <onetick.py.Operation>` are also supported:

        >>> # OTdirective: snippet-name: column operations.is in.from fields;
        >>> data = otp.Ticks(A=["ab", "cv", "bc", "a", "d"], B=["a", "c", "b", "a", "a"])
        >>> yes, no = data[data["A"].isin(data["B"], data["B"] + "b")]
        >>> otp.run(yes)[["A", "B"]]
            A  B
        0  ab  a
        1   a  a
        """
        if items and isinstance(items[0], Sequence) and not isinstance(items[0], str):
            if len(items) > 1:
                raise ValueError("If the first argument of isin() function is a list,"
                                 " it must be the only argument passed to a function.")
            items = items[0]
        return _Operation(_methods.isin, [self, items])

    def fillna(self, value):
        """
        Fill :py:class:`~onetick.py.nan` values with ``value``.

        Parameters
        ----------
        value: float, int, :py:class:`~onetick.py.Operation`
            value to use instead :py:class:`~onetick.py.nan`

        Examples
        --------

        Replace NaN values with a constant:

        >>> data = otp.Ticks({'A': [1, otp.nan, 2]})
        >>> data['A'] = data['A'].fillna(100)   # OTdirective: snippet-name: column operations.fillna;
        >>> otp.run(data)
                             Time      A
        0 2003-12-01 00:00:00.000    1.0
        1 2003-12-01 00:00:00.001  100.0
        2 2003-12-01 00:00:00.002    2.0

        Replace NaN values with a value from the previous tick:

        >>> data = otp.Ticks({'A': [1, otp.nan, 2]})
        >>> data['A'] = data['A'].fillna(data['A'][-1])
        >>> otp.run(data)
                             Time    A
        0 2003-12-01 00:00:00.000  1.0
        1 2003-12-01 00:00:00.001  1.0
        2 2003-12-01 00:00:00.002  2.0
        """
        return _Operation(_methods.fillna, [self, value])

    @property
    def str(self):
        """
        Property that provides access to methods specific to string types.

        See also
        --------
        :py:class:`otp.string <onetick.py.types.string>`
        """

        if issubclass(self.dtype, str):
            from onetick.py.core.column_operations.accessors.str_accessor import _StrAccessor
            return _StrAccessor(self)
        else:
            raise TypeError(".str accessor is available only for string type columns")

    @property
    def dt(self):
        """
        Property that provides access to methods specific to datetime types.

        See also
        --------
        :py:class:`otp.nsectime <onetick.py.types.nsectime>`
        :py:class:`otp.msectime <onetick.py.types.msectime>`
        """

        if issubclass(self.dtype, ott.nsectime) \
                or issubclass(self.dtype, ott.msectime):
            from onetick.py.core.column_operations.accessors.dt_accessor import _DtAccessor
            return _DtAccessor(self)
        else:
            raise TypeError(".dt accessor is available only for datetime type columns")

    @property
    def float(self):
        """
        Property that provides access to
        methods specific to float type.
        """
        if issubclass(self.dtype, float):
            from onetick.py.core.column_operations.accessors.float_accessor import _FloatAccessor
            return _FloatAccessor(self)
        else:
            raise TypeError(".float accessor is available only for float type columns")

    @property
    def decimal(self):
        """
        Property that provides access to methods specific to decimal type.

        See also
        --------
        :py:class:`otp.decimal <onetick.py.types.decimal>`
        """
        if self.dtype is ott.decimal:
            from onetick.py.core.column_operations.accessors.decimal_accessor import _DecimalAccessor
            return _DecimalAccessor(self)
        else:
            raise TypeError(".decimal accessor is available only for decimal type columns")

    def __abs__(self):
        """
        Return the absolute value of float or int column.

        Examples
        --------
        >>> t = otp.Tick(A=-1, B=-2.3)
        >>> t['A'] = abs(t['A'])
        >>> t['B'] = abs(t['B'])
        >>> otp.run(t)[['A', 'B']]
           A    B
        0  1  2.3
        """
        return _Operation(_methods.abs, [self])

    def __round__(self, precision=0):
        """
        Rounds value with specified ``precision``.

        Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
        and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity.

        For values that are exactly half-way between two integers (when the fraction part of value is exactly 0.5),
        the rounding method used here is *upwards*, which returns the bigger number.
        For other rounding methods see :func:`otp.math.round <onetick.py.math.round>` function.

        See also
        --------
        :func:`otp.math.round <onetick.py.math.round>`
        :func:`otp.math.floor <onetick.py.math.floor>`
        :func:`otp.math.ceil <onetick.py.math.ceil>`

        Parameters
        ----------
        precision: int
            Number from -12 to 12.
            Positive precision is precision after the floating point.
            Negative precision is precision before the floating point (and the fraction part is dropped in this case).

        Examples
        --------

        By default the ``precision`` is zero and the number is rounded to the closest integer number
        (and to the bigger number when the fraction part of value is exactly 0.5):

        >>> t = otp.Ticks(A=[-123.45, 123.45, 123.5])
        >>> t['ROUND'] = round(t['A'])
        >>> otp.run(t)
                             Time       A  ROUND
        0 2003-12-01 00:00:00.000 -123.45 -123.0
        1 2003-12-01 00:00:00.001  123.45  123.0
        2 2003-12-01 00:00:00.002  123.50  124.0

        Positive precision truncates to the number of digits *after* floating point:

        >>> t = otp.Ticks(A=[-123.45, 123.45])
        >>> t['ROUND1'] = round(t['A'], 1)
        >>> t['ROUND2'] = round(t['A'], 2)
        >>> otp.run(t)
                             Time       A  ROUND1  ROUND2
        0 2003-12-01 00:00:00.000 -123.45  -123.4 -123.45
        1 2003-12-01 00:00:00.001  123.45   123.5  123.45

        Negative precision truncates to the number of digits *before* floating point
        (and the fraction part is dropped in this case):

        >>> t = otp.Ticks(A=[-123.45, 123.45])
        >>> t['ROUND_M1'] = round(t['A'], -1)
        >>> t['ROUND_M2'] = round(t['A'], -2)
        >>> otp.run(t)
                             Time       A  ROUND_M1  ROUND_M2
        0 2003-12-01 00:00:00.000 -123.45    -120.0    -100.0
        1 2003-12-01 00:00:00.001  123.45     120.0     100.0

        Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
        and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity in all cases:

        >>> t = otp.Ticks(A=[otp.inf, -otp.inf, otp.nan])
        >>> t['ROUND_0'] = round(t['A'])
        >>> t['ROUND_P2'] = round(t['A'], 2)
        >>> t['ROUND_M2'] = round(t['A'], -2)
        >>> otp.run(t)
                             Time    A  ROUND_0  ROUND_P2  ROUND_M2
        0 2003-12-01 00:00:00.000  inf      inf       inf       inf
        1 2003-12-01 00:00:00.001 -inf     -inf      -inf      -inf
        2 2003-12-01 00:00:00.002  NaN      NaN       NaN       NaN

        Returns
        -------
        Operation
        """
        if precision is None:
            precision = 0
        return _Operation(_methods.round, [self, precision])

    def __neg__(self):
        """
        Return the negative value of float or int column.

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3)
        >>> t['A'] = -t['A']
        >>> t['B'] = -t['B']
        >>> otp.run(t)[['A', 'B']]
           A    B
        0 -1 -2.3
        """
        return _Operation(_methods.neg, [self])

    def __add__(self, other):
        """
        Return the sum of column and ``other`` value.

        Parameters
        ----------
        other: int, float, str, :ref:`offset <datetime_offsets>`, :py:class:`onetick.py.Column`

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, C='c', D=otp.datetime(2022, 5, 12))
        >>> t['A'] = t['A'] + t['B']
        >>> t['B'] = t['B'] + 1
        >>> t['C'] = t['C'] + '_suffix'
        >>> t['D'] = t['D'] + otp.Day(1)
        >>> otp.run(t)[['A', 'B', 'C', 'D']]
             A    B         C          D
        0  3.3  3.3  c_suffix 2022-05-13
        """
        return _Operation(_methods.add, [self, other])

    def __radd__(self, other):
        """
        See also
        --------
        __add__

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, C='c', D=otp.datetime(2022, 5, 12))
        >>> t['A'] += t['B']
        >>> t['B'] += 1
        >>> t['C'] += '_suffix'
        >>> t['D'] += otp.Day(1)
        >>> otp.run(t)[['A', 'B', 'C', 'D']]
             A    B         C          D
        0  3.3  3.3  c_suffix 2022-05-13
        """
        return _Operation(_methods.add, [other, self])

    def __sub__(self, other):
        """
        Subtract ``other`` value from column.

        Parameters
        ----------
        other: int, float, :ref:`offset <datetime_offsets>`, :py:class:`onetick.py.Column`

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, D=otp.datetime(2022, 5, 12))
        >>> t['A'] = t['A'] - t['B']
        >>> t['B'] = t['B'] - 1
        >>> t['D'] = t['D'] - otp.Day(1)
        >>> otp.run(t)[['A', 'B', 'D']]
             A    B          D
        0 -1.3  1.3 2022-05-11
        """
        return _Operation(_methods.sub, [self, other])

    def __rsub__(self, other):
        """
        See also
        --------
        __sub__

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, D=otp.datetime(2022, 5, 12))
        >>> t['A'] -= t['B']
        >>> t['B'] -= 1
        >>> t['D'] -= otp.Day(1)
        >>> otp.run(t)[['A', 'B', 'D']]
             A    B          D
        0 -1.3  1.3 2022-05-11
        """
        return _Operation(_methods.sub, [other, self])

    def __mul__(self, other):
        """
        Multiply column by ``other`` value.

        Parameters
        ----------
        other: int, float, str, :py:class:`onetick.py.Column`

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, C='c')
        >>> t['A'] = t['A'] * t['B']
        >>> t['B'] = t['B'] * 2
        >>> t['C'] = t['C'] * 3
        >>> otp.run(t)[['A', 'B', 'C']]
             A    B    C
        0  2.3  4.6  ccc
        """
        return _Operation(_methods.mul, [self, other])

    def __rmul__(self, other):
        """
        See also
        --------
        __mul__

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3, C='c')
        >>> t['A'] *= t['B']
        >>> t['B'] *= 2
        >>> t['C'] *= 3
        >>> otp.run(t)[['A', 'B', 'C']]
             A    B    C
        0  2.3  4.6  ccc
        """
        return _Operation(_methods.mul, [other, self])

    def __truediv__(self, other):
        """
        Divide column by ``other`` value.

        Parameters
        ----------
        other: int, float, :py:class:`onetick.py.Column`

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3)
        >>> t['A'] = t['A'] / t['B']
        >>> t['B'] = t['B'] / 2
        >>> otp.run(t)[['A', 'B']]
                  A     B
        0  0.434783  1.15
        """
        return _Operation(_methods.div, [self, other])

    def __rtruediv__(self, other):
        """
        See also
        --------
        __truediv__

        Examples
        --------
        >>> t = otp.Tick(A=1, B=2.3)
        >>> t['A'] /= t['B']
        >>> t['B'] /= 2
        >>> otp.run(t)[['A', 'B']]
                  A     B
        0  0.434783  1.15
        """
        return _Operation(_methods.div, [other, self])

    def __mod__(self, other):
        """
        Return modulo of division of int column by ``other`` value.

        Parameters
        ----------
        other: int, :py:class:`onetick.py.Column`

        Examples
        --------
        >>> t = otp.Tick(A=3, B=3)
        >>> t['A'] = t['A'] % t['B']
        >>> t['B'] = t['B'] % 2
        >>> otp.run(t)[['A', 'B']]
           A  B
        0  0  1
        """
        return _Operation(_methods.mod, [self, other])

    def __invert__(self):
        """
        Return inversion of filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where(~(t['A'] > 1))
        >>> otp.run(t)[['A']]
           A
        0  0
        1  1
        """
        result = _Operation(_methods.invert, [self])
        return result

    def __eq__(self, other):
        """
        Return equality in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where((t['A'] == 1))
        >>> otp.run(t)[['A']]
           A
        0  1
        """
        result = _Operation(_methods.eq, [self, other])
        return result

    def __ne__(self, other):
        """
        Return inequality in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where((t['A'] != 1))
        >>> otp.run(t)[['A']]
           A
        0  0
        1  2
        2  3
        """
        result = _Operation(_methods.ne, [self, other])
        return result

    def __or__(self, other):
        """
        Return logical ``or`` in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where((t['A'] == 1) | (t['A'] == 2))
        >>> otp.run(t)[['A']]
           A
        0  1
        1  2
        """
        result = _Operation(_methods.or_, [self, other])
        return result

    def __and__(self, other):
        """
        Return logical ``and`` in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=[1, 1], B=[1, 2])
        >>> t = t.where((t['A'] == 1) & (t['B'] == 1))
        >>> otp.run(t)[['A', 'B']]
           A  B
        0  1  1
        """
        result = _Operation(_methods.and_, [self, other])
        return result

    def __le__(self, other):
        """
        Return <= in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where(t['A'] <= 2)
        >>> otp.run(t)[['A']]
           A
        0  0
        1  1
        2  2
        """
        result = _Operation(_methods.le, [self, other])
        return result

    def __lt__(self, other):
        """
        Return < in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where(t['A'] < 2)
        >>> otp.run(t)[['A']]
           A
        0  0
        1  1
        """
        result = _Operation(_methods.lt, [self, other])
        return result

    def __ge__(self, other):
        """
        Return >= in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where(t['A'] >= 2)
        >>> otp.run(t)[['A']]
           A
        0  2
        1  3
        """
        result = _Operation(_methods.ge, [self, other])
        return result

    def __gt__(self, other):
        """
        Return > in filter operation.

        Examples
        --------
        >>> t = otp.Ticks(A=range(4))
        >>> t = t.where(t['A'] > 2)
        >>> otp.run(t)[['A']]
           A
        0  3
        """
        result = _Operation(_methods.gt, [self, other])
        return result

    def __format__(self, format_spec):
        # pylint: disable=E0311
        class_name = self.__class__.__name__
        warnings.warn(
            f"Using `{class_name}` in formatted string literals and `str.format` method is disallowed. "
            f"You can use `otp.format` function for formatting `{class_name}` objects.",
            FutureWarning,
            stacklevel=2,
        )

        return super().__format__(format_spec)

    def _evaluate_func(self, *, set_fields=False):
        if self._op_func:
            op_str, dtype = self._op_func(*self._op_params) if self._op_params else self._op_func()
            if set_fields:
                self._op_str = op_str
                self._dtype = dtype
            return op_str, dtype
        return None

    def _convert_to(self, to_type):
        return _Operation(_methods.CONVERSIONS[self.dtype, to_type], [self])

    def _make_python_way_bool_expression(self):
        dtype = ott.get_object_type(self)
        if dtype is bool:
            return self
        if are_ints_not_time(dtype):
            return _Operation(_methods.ne, (self, 0))
        elif are_time(dtype):
            return _Operation(_methods.ne, (self._convert_to(int), 0))
        elif are_floats(dtype):
            return _Operation(_methods.ne, (self, 0.0))
        elif are_strings(dtype):
            return _Operation(_methods.ne, (self, ""))
        else:
            raise TypeError("Filter expression should return bool, int, float or string")

    def _replace_parameters(self, operation_cb, return_replace_tuples=False):
        """
        Replaces operation parameters (on any nesting level)
        by logic defined by ``operation_cb`` and returns new operation object.

        Parameters
        ----------
        operation_cb: callable
            Callable, with one parameter: current parameter.
            If function returns anything current parameter will be replaced with the result of call
        return_replace_tuples:
            If ``True`` then also returns the list of tuples with old and new parameters.

        Returns
        -------
        result: _Operation or Tuple[_Operation, List[Tuple[_Operation, _Operation]]]
            Returns new operation object with parameters replaced by callback.
            Also may return the list of tuples with old and new parameters.
        """
        self_copy = copy.copy(self)

        replace_tuples = []

        class Node:
            def __init__(self, operation, parent):
                self.operation = operation
                self.parent = parent

        nodes_to_reevaluate = []
        nodes_to_process = [Node(self_copy, None)]

        while nodes_to_process:
            current_node = nodes_to_process.pop()
            current_obj = current_node.operation
            if not getattr(current_obj, '_op_params', None):
                continue

            op_params = current_obj._op_params.copy()

            for i, op in enumerate(current_obj._op_params):
                new_op = operation_cb(op)
                if new_op is not None:
                    replace_tuples.append((op, new_op))
                    op_params[i] = new_op
                    nodes_to_reevaluate.append(Node(new_op, current_node))
                else:
                    if isinstance(op, _Operation):
                        op = copy.copy(op)
                        op_params[i] = op
                        nodes_to_process.append(Node(op, current_node))

            current_obj._op_params = op_params

        for node in reversed(nodes_to_reevaluate):
            parent = node.parent
            while parent is not None:
                parent.operation._evaluate_func(set_fields=True)
                parent = parent.parent

        if return_replace_tuples:
            return self_copy, replace_tuples
        return self_copy


_Operation = Operation  # alias to support backward compatibility


class Raw(Operation):
    """
    Data type representing raw OneTick expression.

    Examples
    --------
    >>> t = otp.Tick(A=1)
    >>> t['A'] = '_TIMEZONE'
    >>> t['B'] = otp.raw('_TIMEZONE', dtype=otp.string[64])
    >>> otp.run(t, timezone='Asia/Yerevan')
            Time          A             B
    0 2003-12-01  _TIMEZONE  Asia/Yerevan
    """
    def __init__(self, raw, dtype):
        if dtype is str:
            warnings.warn(
                f'Be careful, default string length in OneTick is {ott.string.DEFAULT_LENGTH}. '
                "Length of the result raw expression can't be calculated automatically, "
                "so you'd better use onetick.py.string type.",
                stacklevel=2,
            )
        super().__init__(op_str=raw, dtype=dtype)


class OnetickParameter(Operation):
    """
    Data type representing OneTick parameter.

    This object can be used in all places where :class:`~onetick.py.Operation` can be used
    and also in some additional places, e.g. some functions' parameters.

    Parameters
    ----------
    name: str
        The name of the parameter.
    dtype:
        The type of the parameter.
    string_literal: bool
        By default, OneTick inserts string parameters as is in the graph,
        and they are interpreted as an expression.

        In case this parameter is ``True``
        quotes are added when evaluating string parameters, making them the string literals.
        We assume that in most cases it's desirable for the user's parameters to be used as a string literal,
        so the default value for this is ``True``.

        If you want your string parameters to be interpreted as OneTick expressions, set this to ``False``.

    Examples
    --------

    Generate tick with parametrized field value:

    >>> t = otp.Tick(A=otp.param('PARAM', dtype=int))
    >>> otp.run(t, query_params={'PARAM': 1})
            Time  A
    0 2003-12-01  1

    Set bucket interval in aggregation as a parameter:

    >>> t = otp.Ticks(A=[1, 2, 3, 4, 5, 6])
    >>> t = t.agg({'A': otp.agg.sum('A')},
    ...           bucket_units='ticks', bucket_interval=otp.param('PARAM', dtype=int))
    >>> otp.run(t, query_params={'PARAM': 3})
                         Time   A
    0 2003-12-01 00:00:00.002   6
    1 2003-12-01 00:00:00.005  15

    Parameter ``string_literal=False`` can be used if you need string to be interpreted as an expression:

    >>> t = otp.Tick(A=otp.param('PARAM', dtype=str),
    ...              B=otp.param('PARAM', dtype=str, string_literal=False))
    >>> otp.run(t, query_params={'PARAM': 'TOSTRING(NAN())'})
            Time                A    B
    0 2003-12-01  TOSTRING(NAN())  nan
    """
    def __init__(self, name, dtype, string_literal=True):
        self.parameter_expression = f'${name}'

        parameter_expression = self.parameter_expression
        if string_literal and issubclass(dtype, str):
            # OneTick parameters are inserted as is in the graph,
            # so we need to add quotes around string to make it a literal
            # (otherwise it would be interpreted as a column name)
            parameter_expression = ott.value2str(parameter_expression)

        if dtype is float:
            # BDS-93
            # need atof to be able to convert string 'nan' to float
            parameter_expression = f'atof({ott.value2str(parameter_expression)})'

        super().__init__(op_str=parameter_expression, dtype=dtype)
